import math
import os
import pycdlib
from io import BytesIO
from PIL import Image
from hacktools import common


def extractIso(isofile, extractfolder, workfolder=""):
    common.logMessage("Extracting ISO", isofile, "...")
    common.makeFolder(extractfolder)
    iso = pycdlib.PyCdlib()
    iso.open(isofile)
    for dirname, dirlist, filelist in iso.walk(iso_path="/"):
        common.makeFolders(extractfolder + dirname[1:])
        for file in filelist:
            extracted = BytesIO()
            iso.get_file_from_iso_fp(extracted, iso_path=dirname + "/" + file)
            extracted.seek(0)
            with open(extractfolder + dirname[1:] + "/" + file, "wb") as f:
                f.write(extracted.read())
    iso.close()
    if workfolder != "":
        common.copyFolder(extractfolder, workfolder)
    common.logMessage("Done!")


def repackIso(isofile, isopatch, workfolder, patchfile=""):
    common.logMessage("Repacking ISO", isopatch, "...")
    common.copyFile(isofile, isopatch)
    iso = pycdlib.PyCdlib()
    iso.open(isopatch, "r+b")
    files = common.getFiles(workfolder)
    for file in common.showProgress(files):
        filebytes = BytesIO()
        with open(workfolder + file, "rb") as f:
            filebytes.write(f.read())
        filelen = filebytes.tell()
        filebytes.seek(0)
        iso.modify_file_in_place(filebytes, filelen, "/" + file)
    iso.close()
    common.logMessage("Done!")
    # Create xdelta patch
    if patchfile != "":
        common.logMessage("Creating xdelta patch", patchfile, "...")
        xdelta = common.bundledFile("xdelta.exe")
        if not os.path.isfile(xdelta):
            common.logError("xdelta not found")
        else:
            common.execute(xdelta + " -f -e -s {iso} {isopatch} {patch}".format(iso=isofile, isopatch=isopatch, patch=patchfile), False)
            common.logMessage("Done!")


class ELF():
    sections = []
    sectionsdict = {}


class ELFSection():
    name = ""
    nameoff = 0
    type = 0
    flags = 0
    addr = 0
    offset = 0
    size = 0
    link = 0
    info = 0
    addralign = 0
    entsize = 0


def readELF(infile):
    elf = ELF()
    elf.sections = []
    with common.Stream(infile, "rb") as f:
        f.seek(0x20)
        sectionsoff = f.readUInt()
        f.seek(0x2E)
        sectionsize = f.readUShort()
        sectionnum = f.readUShort()
        shstrndx = f.readUShort()
        common.logDebug("sectionsoff:", sectionsoff, "sectionsize:", sectionsize, "sectionnum", sectionnum, "shstrndx", shstrndx)
        # Read section headers
        f.seek(sectionsoff)
        for i in range(sectionnum):
            section = ELFSection()
            section.nameoff = f.readUInt()
            section.type = f.readUInt()
            section.flags = f.readUInt()
            section.addr = f.readUInt()
            section.offset = f.readUInt()
            section.size = f.readUInt()
            section.link = f.readUInt()
            section.info = f.readUInt()
            section.addralign = f.readUInt()
            section.entsize = f.readUInt()
            elf.sections.append(section)
        # Read section names
        for section in elf.sections:
            f.seek(elf.sections[shstrndx].offset + section.nameoff)
            section.name = f.readNullString()
            elf.sectionsdict[section.name] = section
        for i in range(sectionnum):
            common.logDebug(i, vars(elf.sections[i]))
    return elf


def extractBinaryStrings(elf, foundstrings, infile, func, encoding="shift_jis"):
    with common.Stream(infile, "rb") as f:
        rodata = elf.sectionsdict[".rodata"]
        f.seek(rodata.offset)
        while f.tell() < rodata.offset + rodata.size:
            pos = f.tell()
            check = func(f, encoding)
            if check != "":
                if check not in foundstrings:
                    common.logDebug("Found string at", pos)
                    foundstrings.append(check)
                pos = f.tell() - 1
            f.seek(pos + 1)
    return foundstrings


def repackBinaryStrings(elf, section, infile, outfile, detectFunc, writeFunc, encoding="shift_jis"):
    rodata = elf.sectionsdict[".rodata"]
    with common.Stream(infile, "rb") as fi:
        with common.Stream(outfile, "r+b") as fo:
            fi.seek(rodata.offset)
            while fi.tell() < rodata.offset + rodata.size:
                pos = fi.tell()
                check = detectFunc(fi, encoding)
                if check != "":
                    if check in section and section[check][0] != "":
                        common.logDebug("Replacing string at", pos)
                        fo.seek(pos)
                        endpos = fi.tell() - 1
                        newlen = writeFunc(fo, section[check][0], endpos - pos + 1)
                        if newlen < 0:
                            fo.writeZero(1)
                            common.logError("String", section[check][0], "is too long.")
                        else:
                            fo.writeZero(endpos - fo.tell())
                    else:
                        pos = fi.tell() - 1
                fi.seek(pos + 1)


# https://www.psdevwiki.com/ps3/Graphic_Image_Map_(GIM)
class GIM:
    rootoff = 0
    rootsize = 0
    images = []


class GIMImage:
    picoff = 0
    picsize = 0
    imgoff = 0
    imgsize = 0
    imgframeoff = 0
    format = 0
    width = 0
    height = 0
    tiled = 0
    blockedwidth = 0
    blockedheight = 0
    tilewidth = 0
    tileheight = 0
    paloff = 0
    palsize = 0
    palframeoff = 0
    palformat = 0
    palette = []
    colors = []


class GMO:
    size = 0
    names = []
    offsets = []
    gims = []


def readGMO(file):
    gmo = GMO()
    gmo.names = []
    gmo.offsets = []
    gmo.gims = []
    with common.Stream(file, "rb") as f:
        f.seek(16 + 4)
        gmo.size = f.readUInt()
        f.seek(8, 1)
        while f.tell() < gmo.size + 16:
            offset = f.tell()
            id = f.readUShort()
            if id <= 0x06 or (id >= 0x08 and id <= 0x0B):
                blocklen = f.readUShort()
                if id == 0x0A:  # Texture name
                    f.seek(12, 1)
                    texname = f.readNullString()
                    common.logDebug("0x0A at", offset, texname)
                    gmo.names.append(texname)
                f.seek(offset + blocklen)
            elif id == 0x07 or id == 0x0C or (id >= 0x8000 and id < 0x9000):
                f.seek(2, 1)
                blocklen = f.readUInt()
                if id == 0x8013:  # Texture data
                    f.seek(4, 1)
                    gmo.offsets.append(f.tell())
                    common.logDebug("0x8013 at", offset)
                f.seek(offset + blocklen)
            else:
                common.logError("Unknown ID", id, "at", offset)
                break
    for gimoffset in gmo.offsets:
        gmo.gims.append(readGIM(file, gimoffset))
    return gmo


def readGIM(file, start=0):
    gim = GIM()
    gim.images = []
    with common.Stream(file, "rb") as f:
        f.seek(start + 16)
        gim.rootoff = f.tell()
        id = f.readUShort()
        if id != 0x02:
            common.logError("Unexpected id in block 0:", id)
            return None
        f.seek(2, 1)
        gim.rootsize = f.readUInt()
        nextblock = gim.rootoff + f.readUInt()
        image = None
        while nextblock > 0 and nextblock < start + gim.rootsize + 16:
            f.seek(nextblock)
            nextblock, image = readGIMBlock(f, gim, image)
    return gim


def readGIMBlock(f, gim, image):
    offset = f.tell()
    id = f.readUShort()
    f.seek(2, 1)
    if id == 0xFF:
        # Info block
        common.logDebug("GIM 0xFF at", offset)
        return 0, image
    elif id == 0x03:
        # Picture block
        common.logDebug("GIM 0x03 at", offset)
        image = GIMImage()
        image.palette = []
        image.colors = []
        gim.images.append(image)
        image.picoff = offset
        image.picsize = f.readUInt()
        nextblock = f.readUInt()
        common.logDebug("picoff", image.picoff, "picsize", image.picsize)
        return image.picoff + nextblock, image
    elif id == 0x04:
        # Image block
        common.logDebug("GIM 0x04 at", offset)
        image.imgoff = offset
        image.imgsize = f.readUInt()
        nextblock = f.readUInt()
        f.seek(4, 1)
        image.imgframeoff = f.readUShort()
        f.seek(2, 1)
        image.format = f.readUShort()
        if image.format == 0x04:
            image.bpp = 4
        elif image.format == 0x05:
            image.bpp = 8
        elif image.format == 0x03 or image.format == 0x07:
            image.bpp = 32
        else:
            image.bpp = 16
        image.tiled = f.readUShort()
        image.width = f.readUShort()
        image.height = f.readUShort()
        if image.tiled == 0x01:
            image.tilewidth = 0x80 // image.bpp
            image.tileheight = 8
            image.blockedwidth = math.ceil(image.width / image.tilewidth) * image.tilewidth
            image.blockedheight = math.ceil(image.height / image.tileheight) * image.tileheight
        if image.format > 0x05:
            common.logError("Unsupported image format:", image.format)
            return image.imgoff + nextblock, image
        f.seek(image.imgoff + 32 + image.imgframeoff)
        for i in range(image.blockedheight if image.tiled == 0x01 else image.height):
            for j in range(image.blockedwidth if image.tiled == 0x01 else image.width):
                index = 0
                if image.format == 0x04:
                    index = f.readHalf()
                elif image.format == 0x05:
                    index = f.readByte()
                else:
                    index = readColor(f, image.format)
                image.colors.append(index)
        common.logDebug("imgoff", image.imgoff, "imgsize", image.imgsize, "imgframeoff", image.imgframeoff, "format", image.format, "bpp", image.bpp)
        common.logDebug("tiled", image.tiled, "width", image.width, "height", image.height)
        common.logDebug("blockedwidth", image.blockedwidth, "blockedheight", image.blockedheight, "tilewidth", image.tilewidth, "tileheight", image.tileheight)
        return image.imgoff + nextblock, image
    elif id == 0x05:
        # Palette
        common.logDebug("GIM 0x05 at", offset)
        image.paloff = offset
        image.palsize = f.readUInt()
        nextblock = f.readUInt()
        f.seek(4, 1)
        image.palframeoff = f.readUShort()
        f.seek(2, 1)
        image.palformat = f.readUShort()
        f.seek(image.paloff + 32 + image.palframeoff)
        while f.tell() < image.paloff + nextblock:
            image.palette.append(readColor(f, image.palformat))
        common.logDebug("paloff", image.paloff, "palsize", image.palsize, "palframeoff", image.palframeoff)
        common.logDebug("palformat", image.palformat, "length", len(image.palette))
        return image.paloff + nextblock, image
    else:
        common.logWarning("Skipping unknown block at", offset, ":", id)
        f.seek(4, 1)
        return offset + f.readUInt(), image


def writeGIM(file, gim, infile):
    img = Image.open(infile)
    img = img.convert("RGBA")
    pixels = img.load()
    currheight = 0
    with common.Stream(file, "rb+") as f:
        for image in gim.images:
            f.seek(image.imgoff + 32 + image.imgframeoff)
            if image.tiled == 0x00:
                for i in range(image.height):
                    for j in range(image.width):
                        writeGIMPixel(f, image, pixels[j, currheight + i])
            else:
                for blocky in range(image.blockedheight // image.tileheight):
                    for blockx in range(image.blockedwidth // image.tilewidth):
                        for y in range(image.tileheight):
                            for x in range(image.tilewidth):
                                pixelx = blockx * image.tilewidth + x
                                pixely = currheight + blocky * image.tileheight + y
                                if pixelx >= image.width or pixely >= currheight + image.height:
                                    writeGIMPixel(f, image, None)
                                else:
                                    writeGIMPixel(f, image, pixels[pixelx, pixely])
            if len(image.palette) > 0:
                palsize = 5 * (len(image.palette) // 8)
                currheight += max(image.height, palsize)
            else:
                currheight += image.height


def writeGIMPixel(f, image, color):
    if image.format == 0x04 or image.format == 0x05:
        index = common.getPaletteIndex(image.palette, color, False, 0, -1, True, False) if color is not None else 0
        if image.format == 0x04:
            f.writeHalf(index)
        elif image.format == 0x05:
            f.writeByte(index)
    else:
        writeColor(f, image.format, color if color is not None else (0, 0, 0, 0))


def readColor(f, format):
    r, g, b, a = (0, 0, 0, 255)
    if format == 0x00:  # RGBA5650
        color = f.readUShort()
        r = ((color & 0x001F)) << 3
        g = ((color & 0x07E0) >> 5) << 2
        b = ((color & 0xF800) >> 11) << 3
    elif format == 0x01:  # RGBA5551
        color = f.readUShort()
        r = ((color & 0x001F)) << 3
        g = ((color & 0x03E0) >> 5) << 3
        b = ((color & 0x7C00) >> 10) << 3
        a = ((color & 0x8000) >> 15) << 7
        a += 127
    elif format == 0x02:  # RGBA4444
        color = f.readUShort()
        r = ((color & 0x000F)) * 0x11
        g = ((color & 0x00F0) >> 4) * 0x11
        b = ((color & 0x0F00) >> 8) * 0x11
        a = ((color & 0xF000) >> 12) * 0x11
    elif format == 0x03:  # RGBA8888
        color = f.readUInt()
        r = (color & 0x000000FF)
        g = (color & 0x0000FF00) >> 8
        b = (color & 0x00FF0000) >> 16
        a = (color & 0xFF000000) >> 24
    return (r, g, b, a)


def writeColor(f, format, color):
    if format == 0x00:  # RGBA5650
        enc = ((color[2] >> 3) << 11) | ((color[1] >> 2) << 5) | (color[0] >> 3)
        f.writeUShort(enc)
    elif format == 0x01:  # RGBA5551
        a = color[3] - 127
        enc = ((a >> 7) << 15) | ((color[2] >> 3) << 10) | ((color[1] >> 3) << 5) | (color[0] >> 3)
        f.writeUShort(enc)
    elif format == 0x02:  # RGBA4444
        enc = ((color[3] >> 4) << 12) | ((color[2] >> 4) << 8) | ((color[1] >> 4) << 4) | (color[0] >> 4)
        f.writeUShort(enc)
    elif format == 0x03:  # RGBA8888
        enc = (color[3] << 24) | (color[2] << 16) | (color[1] << 8) | color[0]
        f.writeUInt(enc)


def drawGIM(outfile, gim):
    width = 0
    height = 0
    palette = False
    for image in gim.images:
        width = max(width, image.width)
        if len(image.palette) > 0:
            palette = True
            palsize = 5 * (len(image.palette) // 8)
            height += max(image.height, palsize)
        else:
            height += image.height
    img = Image.new("RGBA", (width + (40 if palette else 0), height), (0, 0, 0, 0))
    pixels = img.load()
    currheight = 0
    for image in gim.images:
        i = 0
        if image.tiled == 0x00:
            for y in range(image.height):
                for x in range(image.width):
                    drawGIMPixel(image, pixels, x, currheight + y, i)
                    i += 1
        else:
            for blocky in range(image.blockedheight // image.tileheight):
                for blockx in range(image.blockedwidth // image.tilewidth):
                    for y in range(image.tileheight):
                        for x in range(image.tilewidth):
                            pixelx = blockx * image.tilewidth + x
                            pixely = currheight + blocky * image.tileheight + y
                            if pixelx >= image.width or pixely >= currheight + image.height:
                                i += 1
                                continue
                            drawGIMPixel(image, pixels, pixelx, pixely, i)
                            i += 1
        if len(image.palette) > 0:
            pixels = common.drawPalette(pixels, image.palette, image.width, currheight)
            palsize = 5 * (len(image.palette) // 8)
            currheight += max(image.height, palsize)
        else:
            currheight += image.height
    img.save(outfile, "PNG")


def drawGIMPixel(image, pixels, x, y, i):
    if len(image.palette) > 0:
        pixels[x, y] = image.palette[image.colors[i]]
    else:
        pixels[x, y] = image.colors[i]



import codecs
from enum import IntFlag
import os
import struct
import crcmod
from ndspy import codeCompression
from hacktools import common, compression


def extractRom(romfile, extractfolder, workfolder=""):
    common.logMessage("Extracting ROM", romfile, "...")
    ndstool = common.bundledFile("ndstool.exe")
    if not os.path.isfile(ndstool):
        common.logError("ndstool not found")
    else:
        common.makeFolder(extractfolder)
        common.execute(ndstool + " -x {rom} -9 {folder}arm9.bin -7 {folder}arm7.bin -y9 {folder}y9.bin -y7 {folder}y7.bin -t {folder}banner.bin -h {folder}header.bin -d {folder}data -y {folder}overlay".
                       format(rom=romfile, folder=extractfolder), False)
        if workfolder != "":
            common.copyFolder(extractfolder, workfolder)
        common.logMessage("Done!")


def repackRom(romfile, rompatch, workfolder, patchfile=""):
    common.logMessage("Repacking ROM", rompatch, "...")
    ndstool = common.bundledFile("ndstool.exe")
    if not os.path.isfile(ndstool):
        common.logError("ndstool not found")
    else:
        common.execute(ndstool + " -c {rom} -9 {folder}arm9.bin -7 {folder}arm7.bin -y9 {folder}y9.bin -y7 {folder}y7.bin -t {folder}banner.bin -h {folder}header.bin -d {folder}data -y {folder}overlay".
                       format(rom=rompatch, folder=workfolder), False)
        common.logMessage("Done!")
        # Create xdelta patch
        if patchfile != "":
            common.logMessage("Creating xdelta patch", patchfile, "...")
            xdelta = common.bundledFile("xdelta.exe")
            if not os.path.isfile(xdelta):
                common.logError("xdelta not found")
            else:
                common.execute(xdelta + " -f -e -s {rom} {rompatch} {patch}".format(rom=romfile, rompatch=rompatch, patch=patchfile), False)
                common.logMessage("Done!")


def editBannerTitle(file, title):
    with common.Stream(file, "r+b") as f:
        for i in range(6):
            # Write new text for all languages
            f.seek(576 + 256 * i)
            for char in title:
                f.writeByte(ord(char))
                f.writeByte(0x00)
            # Compute CRC
            f.seek(32)
            crc = crcmod.predefined.mkCrcFun("modbus")(f.read(2080))
            f.seek(2)
            f.writeUShort(crc)


def getHeaderID(file):
    with common.Stream(file, "rb") as f:
        f.seek(12)
        return f.readString(6)


# Binary-related functions
def extractBIN(range, detectFunc=common.detectEncodedString, encoding="shift_jis", binin="data/extract/arm9.bin", binfile="data/bin_output.txt"):
    common.logMessage("Extracting BIN to", binfile, "...")
    foundstrings = extractBinaryStrings(binin, binfile, range, detectFunc, encoding)
    common.logMessage("Done! Extracted", len(foundstrings), "lines")


def extractBinaryStrings(infile, outfile, binrange, func=common.detectEncodedString, encoding="shift_jis", writepos=False):
    foundstrings = []
    insize = os.path.getsize(infile)
    with codecs.open(outfile, "w", "utf-8") as out:
        with common.Stream(infile, "rb") as f:
            f.seek(binrange[0])
            while f.tell() < binrange[1] and f.tell() < insize - 2:
                pos = f.tell()
                check = func(f, encoding)
                if check != "":
                    if check not in foundstrings:
                        common.logDebug("Found string at", pos)
                        foundstrings.append(check)
                        if writepos:
                            out.write(str(pos) + "!")
                        out.write(check + "=\n")
                    pos = f.tell() - 1
                f.seek(pos + 1)
    return foundstrings


def repackBIN(range, freeranges=None, detectFunc=common.detectEncodedString, writeFunc=common.writeEncodedString, encoding="shift_jis", comments="#", binin="data/extract/arm9.bin", binout="data/repack/arm9.bin", binfile="data/bin_input.txt"):
    if not os.path.isfile(binfile):
        common.logError("Input file", binfile, "not found")
        return False

    common.copyFile(binin, binout)
    common.logMessage("Repacking BIN from", binfile, "...")
    section = {}
    with codecs.open(binfile, "r", "utf-8") as bin:
        section = common.getSection(bin, "", comments)
        chartot, transtot = common.getSectionPercentage(section)
    repackBinaryStrings(section, binin, binout, range, freeranges, detectFunc, writeFunc, encoding)
    common.logMessage("Done! Translation is at {0:.2f}%".format((100 * transtot) / chartot))
    return True


def repackBinaryStrings(section, infile, outfile, binrange, freeranges=None, detectFunc=common.detectEncodedString, writeFunc=common.writeEncodedString, encoding="shift_jis"):
    insize = os.path.getsize(infile)
    with common.Stream(infile, "rb") as fi:
        if freeranges is not None:
            allbin = fi.read()
            strpointers = {}
            freeranges = [list(x) for x in freeranges]
        with common.Stream(outfile, "r+b") as fo:
            fi.seek(binrange[0])
            while fi.tell() < binrange[1] and fi.tell() < insize - 2:
                pos = fi.tell()
                check = detectFunc(fi, encoding)
                if check != "":
                    if check in section and section[check][0] != "":
                        common.logDebug("Replacing string at", pos)
                        newsjis = section[check][0]
                        if newsjis == "!":
                            newsjis = ""
                        fo.seek(pos)
                        endpos = fi.tell() - 1
                        newlen = writeFunc(fo, newsjis, endpos - pos + 1, encoding)
                        fo.seek(-1, 1)
                        if fo.readByte() != 0:
                            fo.writeZero(1)
                        if newlen < 0:
                            if freeranges is None:
                                common.logError("String", newsjis, "is too long.")
                            else:
                                range = None
                                rangelen = 0
                                for c in newsjis:
                                    rangelen += 1 if ord(c) < 256 else 2
                                for freerange in freeranges:
                                    if freerange[1] - freerange[0] > rangelen:
                                        range = freerange
                                        break
                                if range is None and newsjis not in strpointers:
                                    common.logWarning("No more room! Skipping", newsjis, "...")
                                else:
                                    # Write the string in a new portion of the rom
                                    if newsjis in strpointers:
                                        newpointer = strpointers[newsjis]
                                    else:
                                        common.logDebug("No room for the string", newsjis, ", redirecting to", common.toHex(range[0]))
                                        fo.seek(range[0])
                                        writeFunc(fo, newsjis, 0, encoding)
                                        fo.seek(-1, 1)
                                        if fo.readByte() != 0:
                                            fo.writeZero(1)
                                        newpointer = 0x02000000 + range[0]
                                        range[0] = fo.tell()
                                        strpointers[newsjis] = newpointer
                                    # Search and replace the old pointer
                                    pointer = 0x02000000 + pos
                                    pointersearch = struct.pack("<I", pointer)
                                    index = 0
                                    common.logDebug("Searching for pointer", common.toHex(pointer))
                                    foundone = False
                                    while index < len(allbin):
                                        index = allbin.find(pointersearch, index)
                                        if index < 0:
                                            break
                                        foundone = True
                                        common.logDebug("Replaced pointer at", str(index))
                                        fo.seek(index)
                                        fo.writeUInt(newpointer)
                                        index += 4
                                    if not foundone:
                                        common.logError("Pointer", common.toHex(pointer), "not found for string", newsjis)
                        else:
                            fo.writeZero(endpos - fo.tell())
                    else:
                        pos = fi.tell() - 1
                fi.seek(pos + 1)


# Compression-related functions
class CompressionType(IntFlag):
    LZ10 = 0x10,
    LZ11 = 0x11,
    Huff4 = 0x24,
    Huff8 = 0x28,
    RLE = 0x30,
    LZ40 = 0x40,
    LZ60 = 0x60


def decompress(f, complength):
    header = f.readUInt()
    type = header & 0xFF
    decomplength = ((header & 0xFFFFFF00) >> 8)
    common.logDebug("Compression header:", common.toHex(header), "type:", common.toHex(type), "length:", decomplength)
    with common.Stream() as data:
        data.write(f.read(complength - 4))
        data.seek(0)
        if type == CompressionType.LZ10:
            return compression.decompressLZ10(data, complength, decomplength)
        elif type == CompressionType.LZ11:
            return compression.decompressLZ11(data, complength, decomplength)
        else:
            common.logError("Unsupported compression type", common.toHex(type))
            return data.read()


def compress(data, type):
    with common.Stream() as out:
        length = len(data)
        out.writeByte(type.value)
        out.writeByte(length & 0xFF)
        out.writeByte(length >> 8 & 0xFF)
        out.writeByte(length >> 16 & 0xFF)
        if type == CompressionType.LZ10:
            out.write(compression.compressLZ10(data))
        elif type == CompressionType.LZ11:
            out.write(compression.compressLZ11(data))
        else:
            common.logError("Unsupported compression type", common.toHex(type))
            out.write(data)
        out.seek(0)
        return out.read()


def decompressFile(infile, outfile):
    insize = os.path.getsize(infile)
    with common.Stream(infile, "rb") as fin:
        with common.Stream(outfile, "wb") as fout:
            fout.write(decompress(fin, insize - 4))


def compressFile(infile, outfile, type):
    with common.Stream(infile, "rb") as fin:
        data = fin.read()
        with common.Stream(outfile, "wb") as fout:
            fout.write(compress(data, type))


def decompressBinary(infile, outfile):
    with common.Stream(infile, "rb") as fin:
        data = fin.read()
    uncdata = codeCompression.decompress(data)
    with common.Stream(outfile, "wb") as f:
        f.write(uncdata)


def compressBinary(infile, outfile):
    with common.Stream(infile, "rb") as fin:
        data = bytearray(fin.read())
    compdata = bytearray(codeCompression.compress(data, True))
    codeoffset = 0
    for i in range(0, 0x8000, 4):
        if compdata[i:i+8] == b'\x21\x06\xC0\xDE\xDE\xC0\x06\x21':
            codeoffset = i - 0x1C
            break
    if codeoffset > 0:
        struct.pack_into("<I", compdata, codeoffset + 0x14, 0x02000000 + len(compdata))
    with common.Stream(outfile, "wb") as f:
        f.write(compdata)



import ctypes
from hacktools import common


# Implementations based on Kurimuu's Kontract
# https://github.com/IcySon55/Kuriimu/tree/master/src/Kontract/Compression
def decompressLZ10(data, complength, decomplength, dispextra=1):
    with common.Stream() as out:
        readbytes = 0

        # the maximum 'DISP-1' is 0xFFF.
        bufferlength = 0x1000
        buffer = bytearray(bufferlength)
        bufferoffset = 0

        currentoutsize = 0
        flags = 0
        mask = 1
        while currentoutsize < decomplength:
            # Update the mask. If all flag bits have been read, get a new set.
            # the current mask is the mask used in the previous run. So if it masks the
            # last flag bit, get a new flags byte.
            if mask == 1:
                if readbytes >= complength:
                    raise Exception("Not enough data.")
                flags = data.readByte()
                readbytes += 1
                if flags < 0:
                    raise Exception("Stream too short!")
                mask = 0x80
            else:
                mask >>= 1
            # bit = 1 <=> compressed.
            if (flags & mask) > 0:
                # Get length and displacement('disp') values from next 2 bytes
                # there are < 2 bytes available when the end is at most 1 byte away
                if readbytes + 1 >= complength:
                    if readbytes < complength:
                        data.readbyte()
                        readbytes += 1
                    raise Exception("Not enough data.")
                byte1 = data.readByte()
                byte2 = data.readByte()
                readbytes += 2
                if byte2 < 0:
                    raise Exception("Stream too short!")
                # the number of bytes to copy
                length = byte1 >> 4
                length += 3
                # from where the bytes should be copied (relatively)
                disp = ((byte1 & 0x0F) << 8) | byte2
                disp += dispextra
                if disp > currentoutsize:
                    raise Exception("Cannot go back more than already written.")

                bufidx = bufferoffset + bufferlength - disp
                for i in range(length):
                    next = buffer[bufidx % bufferlength]
                    bufidx += 1
                    out.writeByte(next)
                    buffer[bufferoffset] = next
                    bufferoffset = (bufferoffset + 1) % bufferlength
                currentoutsize += length
            else:
                if readbytes >= complength:
                    raise Exception("Not enough data.")
                next = data.readByte()
                readbytes += 1
                if next < 0:
                    raise Exception("Stream too short!")
                currentoutsize += 1
                out.writeByte(next)
                buffer[bufferoffset] = next
                bufferoffset = (bufferoffset + 1) % bufferlength
        out.seek(0)
        return out.read(decomplength)


def compressLZ10(indata, mindisp=1):
    with common.Stream() as out:
        inlength = len(indata)
        compressedlength = 0
        instart = 0
        # we do need to buffer the output, as the first byte indicates which blocks are compressed.
        # this version does not use a look-ahead, so we do not need to buffer more than 8 blocks at a time.
        outbuffer = bytearray(8 * 2 + 1)
        outbuffer[0] = 0
        bufferlength = 1
        bufferedblocks = 0
        readbytes = 0
        while readbytes < inlength:
            # If 8 blocks are buffered, write them and reset the buffer
            # we can only buffer 8 blocks at a time.
            if bufferedblocks == 8:
                out.write(outbuffer[:bufferlength])
                compressedlength += bufferlength
                # reset the buffer
                outbuffer[0] = 0
                bufferlength = 1
                bufferedblocks = 0
            # determine if we're dealing with a compressed or raw block.
            # it is a compressed block when the next 3 or more bytes can be copied from
            # somewhere in the set of already compressed bytes.
            oldlength = min(readbytes, 0x1000)
            length, disp = getOccurrenceLength(indata, instart + readbytes, min(inlength - readbytes, 0x12), instart + readbytes - oldlength, oldlength, mindisp)
            # length not 3 or more? next byte is raw data
            if length < 3:
                outbuffer[bufferlength] = indata[instart + readbytes]
                bufferlength += 1
                readbytes += 1
            else:
                # 3 or more bytes can be copied? next (length) bytes will be compressed into 2 bytes
                readbytes += length
                # mark the next block as compressed
                outbuffer[0] |= (1 << (7 - bufferedblocks))
                outbuffer[bufferlength] = ((length - 3) << 4) & 0xF0
                outbuffer[bufferlength] |= ((disp - 1) >> 8) & 0x0F
                bufferlength += 1
                outbuffer[bufferlength] = (disp - 1) & 0xFF
                bufferlength += 1
            bufferedblocks += 1
        # copy the remaining blocks to the output
        if bufferedblocks > 0:
            out.write(outbuffer[:bufferlength])
            compressedlength += bufferlength
        out.seek(0)
        return out.read()


def decompressLZ11(data, complength, decomplength):
    result = bytearray(decomplength)
    dstoffset = 0
    while True:
        header = data.readByte()
        for i in range(8):
            if (header & 0x80) == 0:
                result[dstoffset] = data.readByte()
                dstoffset += 1
            else:
                a = data.readByte()
                b = data.readByte()
                offset = 0
                length2 = 0
                if (a >> 4) == 0:
                    c = data.readByte()
                    length2 = (((a & 0xF) << 4) | (b >> 4)) + 0x11
                    offset = (((b & 0xF) << 8) | c) + 1
                elif (a >> 4) == 1:
                    c = data.readByte()
                    d = data.readByte()
                    length2 = (((a & 0xF) << 12) | (b << 4) | (c >> 4)) + 0x111
                    offset = (((c & 0xF) << 8) | d) + 1
                else:
                    length2 = (a >> 4) + 1
                    offset = (((a & 0xF) << 8) | b) + 1
                for j in range(length2):
                    result[dstoffset] = result[dstoffset - offset]
                    dstoffset += 1
            if dstoffset >= decomplength:
                return result
            header <<= 1


def compressLZ11(indata, mindisp=1):
    with common.Stream() as out:
        inlength = len(indata)
        compressedlength = 0
        instart = 0
        # we do need to buffer the output, as the first byte indicates which blocks are compressed.
        # this version does not use a look-ahead, so we do not need to buffer more than 8 blocks at a time.
        # (a block is at most 4 bytes long)
        outbuffer = bytearray(8 * 4 + 1)
        outbuffer[0] = 0
        bufferlength = 1
        bufferedblocks = 0
        readbytes = 0
        while readbytes < inlength:
            # If 8 blocks are buffered, write them and reset the buffer
            # we can only buffer 8 blocks at a time.
            if bufferedblocks == 8:
                out.write(outbuffer[:bufferlength])
                compressedlength += bufferlength
                # reset the buffer
                outbuffer[0] = 0
                bufferlength = 1
                bufferedblocks = 0
            # determine if we're dealing with a compressed or raw block.
            # it is a compressed block when the next 3 or more bytes can be copied from
            # somewhere in the set of already compressed bytes.
            oldlength = min(readbytes, 0x1000)
            length, disp = getOccurrenceLength(indata, instart + readbytes, min(inlength - readbytes, 0x10110), instart + readbytes - oldlength, oldlength, mindisp)
            # length not 3 or more? next byte is raw data
            if length < 3:
                outbuffer[bufferlength] = indata[instart + readbytes]
                bufferlength += 1
                readbytes += 1
            else:
                # 3 or more bytes can be copied? next (length) bytes will be compressed into 2 bytes
                readbytes += length
                # mark the next block as compressed
                outbuffer[0] |= (1 << (7 - bufferedblocks))
                if length >= 0x110:
                    # case 1: 1(B CD E)(F GH) + (0x111)(0x1) = (LEN)(DISP)
                    outbuffer[bufferlength] = 0x10
                    outbuffer[bufferlength] |= ((length - 0x111) >> 12) & 0x0F
                    bufferlength += 1
                    outbuffer[bufferlength] = ((length - 0x111) >> 4) & 0xFF
                    bufferlength += 1
                    outbuffer[bufferlength] = ((length - 0x111) << 4) & 0xF0
                elif length > 0x10:
                    # case 0; 0(B C)(D EF) + (0x11)(0x1) = (LEN)(DISP)
                    outbuffer[bufferlength] = 0x00
                    outbuffer[bufferlength] |= ((length - 0x11) >> 4) & 0x0F
                    bufferlength += 1
                    outbuffer[bufferlength] = ((length - 0x11) << 4) & 0xF0
                else:
                    # case > 1: (A)(B CD) + (0x1)(0x1) = (LEN)(DISP)
                    outbuffer[bufferlength] = ((length - 1) << 4) & 0xF0
                # the last 1.5 bytes are always the disp
                outbuffer[bufferlength] |= ((disp - 1) >> 8) & 0x0F
                bufferlength += 1
                outbuffer[bufferlength] = (disp - 1) & 0xFF
                bufferlength += 1
            bufferedblocks += 1
        # copy the remaining blocks to the output
        if bufferedblocks > 0:
            out.write(outbuffer[:bufferlength])
            compressedlength += bufferlength
        out.seek(0)
        return out.read()


def getOccurrenceLength(indata, newptr, newlength, oldptr, oldlength, mindisp=1):
    disp = 0
    if newlength == 0:
        return 0
    maxlength = 0
    # try every possible 'disp' value (disp = oldLength - i)
    for i in range(oldlength - mindisp):
        # work from the start of the old data to the end, to mimic the original implementation's behaviour
        # (and going from start to end or from end to start does not influence the compression ratio anyway)
        currentoldstart = oldptr + i
        currentlength = 0
        # determine the length we can copy if we go back (oldLength - i) bytes
        # always check the next 'newLength' bytes, and not just the available 'old' bytes,
        # as the copied data can also originate from what we're currently trying to compress.
        for j in range(newlength):
            # stop when the bytes are no longer the same
            if indata[currentoldstart + j] != indata[newptr + j]:
                break
            currentlength += 1
        # update the optimal value
        if currentlength > maxlength:
            maxlength = currentlength
            disp = oldlength - i
            # if we cannot do better anyway, stop trying.
            if maxlength == newlength:
                break
    return maxlength, disp


# https://forum.xentax.com/viewtopic.php?p=30390#p30387
def getBits(n, f, blen, fbuf):
    retv = 0
    while n > 0:
        retv = retv << 1
        if blen == 0:
            fbuf = f.readSByte()
            blen = 8
        if fbuf & 0x80:
            retv |= 1
        fbuf = fbuf << 1
        blen -= 1
        n -= 1
    return retv, blen, fbuf


def decompressPRS(f, slen, dlen):
    dbuf = bytearray(dlen)
    startpos = f.tell()
    blen = 0
    fbuf = 0
    dptr = 0
    len = 0
    pos = 0
    while f.tell() < startpos + slen:
        flag, blen, fbuf = getBits(1, f, blen, fbuf)
        if flag == 1:
            if dptr < dlen:
                dbuf[dptr] = f.readByte()
                dptr += 1
        else:
            flag, blen, fbuf = getBits(1, f, blen, fbuf)
            if flag == 0:
                len, blen, fbuf = getBits(2, f, blen, fbuf)
                len += 2
                data = f.readSByte()
                # Use ctypes to correctly handle int overflow
                pos = ctypes.c_int(data | 0xffffff00).value
            else:
                pos = ctypes.c_int((f.readSByte() << 8) | 0xffff0000).value
                pos |= f.readSByte() & 0xff
                len = pos & 0x07
                pos >>= 3
                if len == 0:
                    len = (f.readSByte() & 0xff) + 1
                else:
                    len += 2
            pos += dptr
            for i in range(len):
                if dptr < dlen:
                    dbuf[dptr] = dbuf[pos]
                    dptr += 1
                    pos += 1
    return dbuf



import os
from hacktools import common


class ARCHArchive:
    filenum = 0
    tableoff = 0
    fatoff = 0
    nameindexoff = 0
    dataoff = 0
    files = []


class ARCHFile:
    name = ""
    length = 0
    declength = 0
    offset = 0
    nameoffset = 0
    encoded = False


def read(f):
    f.seek(4)  # Magic: ARCH
    archive = ARCHArchive()
    archive.files = []
    archive.filenum = f.readUInt()
    archive.tableoff = f.readUInt()
    archive.fatoff = f.readUInt()
    archive.nameindexoff = f.readUInt()
    archive.dataoff = f.readUInt()
    common.logDebug("Archive:", vars(archive))
    for i in range(archive.filenum):
        f.seek(archive.fatoff + i * 16)
        subfile = ARCHFile()
        subfile.length = f.readUInt()
        subfile.declength = f.readUInt()
        subfile.offset = f.readUInt()
        subfile.nameoffset = f.readUShort()
        subfile.encoded = f.readUShort() == 1
        f.seek(archive.tableoff + subfile.nameoffset)
        subfile.name = f.readNullString()
        common.logDebug("File", i, vars(subfile))
        archive.files.append(subfile)
    return archive


def repack(fin, f, archive, infolder):
    # Copy everything up to dataoff
    fin.seek(0)
    f.seek(0)
    f.write(fin.read(archive.dataoff))
    # Loop the files
    dataoff = 0
    for i in range(archive.filenum):
        subfile = archive.files[i]
        filepath = infolder + subfile.name
        if not os.path.isfile(filepath):
            # Just update the offset and copy the file
            f.seek(archive.fatoff + i * 16)
            f.seek(8, 1)
            f.writeUInt(dataoff)
            fin.seek(archive.dataoff + subfile.offset)
            f.seek(archive.dataoff + dataoff)
            f.write(fin.read(subfile.length))
        else:
            # Set the file as not encoded and copy it
            size = os.path.getsize(filepath)
            size += size % 16
            f.seek(archive.fatoff + i * 16)
            f.writeUInt(size)
            f.writeUInt(size)
            f.writeUInt(dataoff)
            f.seek(2, 1)
            f.writeUShort(0)
            f.seek(archive.dataoff + dataoff)
            with common.Stream(filepath, "rb") as subf:
                f.write(subf.read())
        # Align with 0s
        if f.tell() % 16 > 0:
            f.writeZero(16 - (f.tell() % 16))
        dataoff = f.tell() - archive.dataoff


def extract(f, archive, outfolder):
    for subfile in archive.files:
        with common.Stream(outfolder + subfile.name, "wb") as fout:
            f.seek(archive.dataoff + subfile.offset)
            if not subfile.encoded:
                fout.write(f.read(subfile.length))
            else:
                # Based on Tinke's ARCH implementation
                startpos = f.tell()
                buffer1 = []
                buffer2 = []
                for i in range(0x100):
                    buffer1.append(0)
                    buffer2.append(0)
                while f.tell() - startpos < subfile.length:
                    # InitBuffer
                    for i in range(0x100):
                        buffer2[i] = i
                    # FillBuffer
                    index = 0
                    while index != 0x100:
                        id = f.readByte()
                        numloops = id
                        if id > 0x7F:
                            numloops = 0
                            index += id - 0x7F
                        if index == 0x100:
                            break
                        if numloops < 0:
                            continue
                        for i in range(numloops + 1):
                            byte = f.readByte()
                            buffer2[index] = byte
                            if byte != index:
                                buffer1[index] = f.readByte()
                            index += 1
                    # Process
                    numloops = (f.readByte() << 8) + f.readByte()
                    nextsamples = []
                    while True:
                        if len(nextsamples) == 0:
                            if numloops == 0:
                                break
                            numloops -= 1
                            index = f.readByte()
                        else:
                            index = nextsamples.pop()
                        if buffer2[index] == index:
                            fout.writeByte(index)
                        else:
                            nextsamples.append(buffer1[index])
                            nextsamples.append(buffer2[index])
                            index = len(nextsamples)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="hacktools",
    version="0.10.3",
    author="Illidan",
    description="A set of utilities and tools for rom hacking and translations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Illidanz/hacktools",
    packages=["hacktools"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "click",
        "tqdm",
        "ndspy",
        "crcmod",
        "Pillow",
        "pycdlib",
        "psd-tools>=1.8,<1.9"
    ],
    python_requires=">=3.7",
)



import codecs
from io import BytesIO
import distutils.dir_util
import logging
import math
import os
import re
import shutil
import sys
import struct
import subprocess
import click
from tqdm import tqdm

table = {}


# File reading
class Stream(object):
    def __init__(self, fpath="", mode="m", little=True):
        self.f = fpath
        self.mode = mode
        self.endian = "<" if little else ">"
        self.half = None

    def __enter__(self):
        if self.mode == "m":
            self.f = BytesIO()
        else:
            self.f = open(self.f, self.mode)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.f.close()

    def tell(self):
        return self.f.tell()

    def seek(self, pos, whence=0):
        self.f.seek(pos, whence)

    def read(self, n=-1):
        return self.f.read(n)

    def write(self, data):
        self.f.write(data)

    def peek(self, n):
        pos = self.tell()
        ret = self.read(n)
        self.seek(pos)
        return ret

    def writeLine(self, data):
        self.f.write(data + "\n")

    def readInt(self):
        return struct.unpack(self.endian + "i", self.read(4))[0]

    def readUInt(self):
        return struct.unpack(self.endian + "I", self.read(4))[0]

    def readShort(self):
        return struct.unpack(self.endian + "h", self.read(2))[0]

    def readUShort(self):
        return struct.unpack(self.endian + "H", self.read(2))[0]

    def readByte(self):
        return struct.unpack("B", self.read(1))[0]

    def readSByte(self):
        return struct.unpack("b", self.read(1))[0]

    def readHalf(self):
        if self.half is None:
            self.half = self.readByte()
            return self.half & 0x0F
        else:
            ret = self.half >> 4
            self.half = None
            return ret

    def readZeros(self, size):
        while self.tell() < size:
            byte = self.readByte()
            if byte != 0x00:
                self.seek(-1, 1)
                break

    def readBytes(self, n):
        ret = ""
        for i in range(n):
            ret += toHex(self.readByte()) + " "
        return ret

    def readString(self, length):
        str = ""
        for i in range(length):
            byte = self.readByte()
            # These control characters can be found in texture names, replace them with a space
            if byte == 0x82 or byte == 0x86:
                byte = 0x20
            if byte != 0:
                str += chr(byte)
        return str

    def readNullString(self):
        str = ""
        while True:
            byte = self.readByte()
            if byte == 0:
                break
            else:
                str += chr(byte)
        return str

    def writeInt(self, num):
        self.f.write(struct.pack(self.endian + "i", num))

    def writeIntAt(self, pos, num):
        current = self.tell()
        self.seek(pos)
        self.writeInt(num)
        self.seek(current)

    def writeUInt(self, num):
        self.f.write(struct.pack(self.endian + "I", num))

    def writeUIntAt(self, pos, num):
        current = self.tell()
        self.seek(pos)
        self.writeUInt(num)
        self.seek(current)

    def writeShort(self, num):
        self.f.write(struct.pack(self.endian + "h", num))

    def writeShortAt(self, pos, num):
        current = self.tell()
        self.seek(pos)
        self.writeShort(num)
        self.seek(current)

    def writeUShort(self, num):
        self.f.write(struct.pack(self.endian + "H", num))

    def writeUShortAt(self, pos, num):
        current = self.tell()
        self.seek(pos)
        self.writeUShort(num)
        self.seek(current)

    def writeByte(self, num):
        self.f.write(struct.pack("B", num))

    def writeByteAt(self, pos, num):
        current = self.tell()
        self.seek(pos)
        self.writeByte(num)
        self.seek(current)

    def writeSByte(self, num):
        self.f.write(struct.pack("b", num))

    def writeSByteAt(self, pos, num):
        current = self.tell()
        self.seek(pos)
        self.writeSByte(num)
        self.seek(current)

    def writeHalf(self, num):
        if self.half is None:
            self.half = num
        else:
            self.writeByte((num << 4) | self.half)
            self.half = None

    def writeString(self, str):
        self.f.write(str.encode("ascii"))

    def writeZero(self, num):
        if num > 0:
            for i in range(num):
                self.writeByte(0)


# Logging
@click.group(no_args_is_help=True)
@click.option("--log", is_flag=True, default=False)
def cli(log):
    if log:
        logging.basicConfig(filename="tool.log", filemode="w", format="[%(levelname)s] %(message)s", level=logging.DEBUG)
    else:
        logging.getLogger().disabled = True


def logMessage(*messages):
    message = " ".join(str(x) for x in messages)
    logging.info(message)
    tqdm.write(message)


def logDebug(*messages):
    message = " ".join(str(x) for x in messages)
    logging.debug(message)


def logWarning(*messages):
    message = " ".join(str(x) for x in messages)
    logging.warning(message)


def logError(*messages):
    message = " ".join(str(x) for x in messages)
    logging.error(message)
    tqdm.write("[ERROR] " + message)


def showProgress(iterable):
    return tqdm(iterable=iterable)


# Strings
def toHex(byte):
    hexstr = hex(byte)[2:].upper()
    if len(hexstr) == 1:
        return "0" + hexstr
    return hexstr


def isAscii(s):
    for i in range(len(s)):
        if ord(s[i]) >= 128:
            return False
    return True


def codeToChar(code):
    try:
        if code < 256:
            return struct.pack("B", code).decode("ascii")
        return struct.pack(">H", code).decode("shift_jis")
    except UnicodeDecodeError:
        return ""


def loadTable(tablefile):
    if os.path.isfile(tablefile):
        with codecs.open(tablefile, "r", "utf-8") as ft:
            for line in ft:
                line = line.strip("\r\n")
                if line.find("=") > 0:
                    linesplit = line.split("=", 1)
                    table[linesplit[0]] = linesplit[1]


def shiftPointer(pointer, pointerdiff):
    newpointer = pointer
    for k, v in pointerdiff.items():
        if k < pointer:
            newpointer += v
    if newpointer != pointer:
        logDebug("Shifted pointer", pointer, "to", newpointer)
    return newpointer


def checkShiftJIS(first, second):
    # Based on https://www.lemoda.net/c/detect-shift-jis/
    status = False
    if (first >= 0x81 and first <= 0x84) or (first >= 0x87 and first <= 0x9f):
        if second >= 0x40 and second <= 0xfc:
            status = True
    elif first >= 0xe0 and first <= 0xef:
        if second >= 0x40 and second <= 0xfc:
            status = True
    return status


def getSection(f, title, comment="#", fixchars=[]):
    ret = {}
    found = title == ""
    try:
        f.seek(0)
        for line in f:
            line = line.rstrip("\r\n").replace("\ufeff", "")
            if not found and line.startswith("!FILE:" + title):
                found = True
            elif found:
                if title != "" and line.startswith("!FILE:"):
                    break
                elif line.find("=") > 0:
                    split = line.split("=", 1)
                    split[1] = split[1].split(comment)[0]
                    if split[0] not in ret:
                        ret[split[0]] = []
                    for fixchar in fixchars:
                        split[1] = split[1].replace(fixchar[0], fixchar[1])
                    ret[split[0]].append(split[1])
    except UnicodeDecodeError:
        return ret
    return ret


def getSectionPercentage(section, chartot=0, transtot=0):
    for s in section.keys():
        strlen = len(s)
        for s2 in section[s]:
            chartot += strlen
            if s2 != "":
                transtot += strlen
    return chartot, transtot


def wordwrap(text, glyphs, width, codefunc=None, default=6, linebreak="|", sectionsep=">>"):
    # Based on http://code.activestate.com/recipes/577946-word-wrap-for-proportional-fonts/
    lines = []
    if text.count(sectionsep) > 0:
        lines = text.split(sectionsep)
        for i in range(len(lines)):
            lines[i] = wordwrap(lines[i], glyphs, width, codefunc, default, linebreak, sectionsep)
        return sectionsep.join(lines)
    text = text.replace(linebreak, "\n")
    pattern = re.compile(r"(\s+)")
    lookup = dict((c, glyphs[c].length if c in glyphs else default) for c in set(text))
    for line in text.splitlines():
        tokens = pattern.split(line)
        tokens.append("")
        widths = []
        for token in tokens:
            tokenwidth = 0
            i = 0
            while i < len(token):
                if codefunc is not None:
                    skip = codefunc(token, i)
                    if skip > 0:
                        i += skip
                        continue
                tokenwidth += lookup[token[i]]
                i += 1
            widths.append(tokenwidth)
        start, total = 0, 0
        for index in range(0, len(tokens), 2):
            if total + widths[index] > width:
                end = index + 2 if index == start else index
                lines.append("".join(tokens[start:end]))
                start, total = end, 0
                if end == index + 2:
                    continue
            total += widths[index] + widths[index + 1]
        if start < len(tokens):
            lines.append("".join(tokens[start:]))
    lines = [line.strip() for line in lines]
    return linebreak.join(lines)


def readEncodedString(f, encoding="shift_jis"):
    sjis = ""
    while True:
        b1 = f.readByte()
        if b1 == 0x0A:
            sjis += "|"
        elif b1 == 0x00:
            break
        else:
            b2 = f.readByte()
            if not checkShiftJIS(b1, b2):
                if b2 == 0x01:
                    sjis += "UNK(" + toHex(b1) + toHex(b2) + ")"
                else:
                    f.seek(-1, 1)
                    sjis += chr(b1)
            else:
                f.seek(-2, 1)
                try:
                    sjis += f.read(2).decode(encoding).replace("〜", "～")
                except UnicodeDecodeError:
                    logError("[ERROR] UnicodeDecodeError")
                    sjis += "[ERROR" + str(f.tell() - 2) + "]"
    return sjis


def detectEncodedString(f, encoding="shift_jis"):
    ret = ""
    sjis = 0
    while True:
        b1 = f.readByte()
        if b1 == 0x0A:
            ret += "|"
        elif b1 == 0x00:
            break
        elif b1 >= 28 and b1 <= 126 and (len(ret) > 0 or b1 == 0x25):
            ret += chr(b1)
        else:
            b2 = f.readByte()
            if checkShiftJIS(b1, b2):
                f.seek(-2, 1)
                try:
                    ret += f.read(2).decode(encoding).replace("〜", "～")
                    sjis += 1
                except UnicodeDecodeError:
                    if ret.count("UNK(") >= 5:
                        return ""
                    ret += "UNK(" + toHex(b1) + toHex(b2) + ")"
            elif len(ret) > 1 and ret.count("UNK(") < 5:
                ret += "UNK(" + toHex(b1) + toHex(b2) + ")"
            else:
                return ""
    return ret


def writeEncodedString(f, s, maxlen=0, encoding="shift_jis"):
    i = 0
    x = 0
    s = s.replace("～", "〜")
    while x < len(s):
        c = s[x]
        if c == "U" and x < len(s) - 4 and s[x:x+4] == "UNK(":
            if maxlen > 0 and i+2 > maxlen:
                return -1
            code = s[x+4] + s[x+5]
            f.write(bytes.fromhex(code))
            code = s[x+6] + s[x+7]
            f.write(bytes.fromhex(code))
            x += 8
            i += 2
        elif c == "|":
            if maxlen > 0 and i+1 > maxlen:
                return -1
            f.writeByte(0x0A)
            i += 1
        elif ord(c) < 128:
            if maxlen > 0 and i+1 > maxlen:
                return -1
            f.writeByte(ord(c))
            i += 1
        else:
            if maxlen > 0 and i+2 > maxlen:
                return -1
            f.write(c.encode(encoding))
            i += 2
        x += 1
    f.writeByte(0x00)
    return i


# Folders
def makeFolder(folder, clear=True):
    if clear:
        clearFolder(folder)
    os.mkdir(folder)


def clearFolder(folder):
    if os.path.isdir(folder):
        shutil.rmtree(folder)


def copyFolder(f1, f2):
    clearFolder(f2)
    shutil.copytree(f1, f2)


def mergeFolder(f1, f2):
    distutils.dir_util.copy_tree(f1, f2)


def copyFile(f1, f2):
    if os.path.isfile(f2):
        os.remove(f2)
    shutil.copyfile(f1, f2)


def makeFolders(path):
    try:
        os.makedirs(path)
    except FileExistsError:
        pass


def getFiles(path, extensions=[]):
    if isinstance(extensions, str) and extensions != "":
        extensions = [extensions]
    ret = []
    for (root, dirs, files) in os.walk(path):
        for file in files:
            file = os.path.join(root, file).replace(path, "").replace("\\", "/")
            if len(extensions) > 0 and os.path.splitext(file)[1] not in extensions:
                continue
            ret.append(file)
    return ret


def bundledFile(name):
    try:
        return os.path.join(sys._MEIPASS, name)
    except AttributeError:
        return name


def execute(cmd, show=True):
    try:
        result = str(subprocess.check_output(cmd))
    except FileNotFoundError:
        logError("Command too long:", len(cmd))
        return
    except subprocess.CalledProcessError:
        logError("Command error", cmd)
        return
    if result != "":
        if show:
            logMessage(result)
        else:
            logDebug(result)


def armipsPatch(file):
    logMessage("Applying armips patch ...")
    armips = bundledFile("armips.exe")
    if not os.path.isfile(armips):
        logError("armips not found")
    else:
        execute(armips + " {binpatch}".format(binpatch=file), False)
        logMessage("Done!")


# Generic texture
def readPalette(p):
    return (((p >> 0) & 0x1f) << 3, ((p >> 5) & 0x1f) << 3, ((p >> 10) & 0x1f) << 3, 0xff)


def getColorDistance(c1, c2, checkalpha=False):
    (r1, g1, b1, a1) = c1
    (r2, g2, b2, a2) = c2
    sum = (r1 - r2) ** 2 + (g1 - g2) ** 2 + (b1 - b2) ** 2
    if checkalpha:
        sum += (a1 - a2) ** 2
    return math.sqrt(sum)


def sumColors(c1, c2, a=1, b=1, c=2):
    (r1, g1, b1, a1) = c1
    (r2, g2, b2, a2) = c2
    return ((r1 * a + r2 * b) // c, (g1 * a + g2 * b) // c, (b1 * a + b2 * b) // c, a1)


# https://github.com/marco-calautti/Rainbow/blob/master/Rainbow.ImgLib/ImgLib/Common/ImageUtils.cs
cc38 = [
    0x00, 0x24, 0x49, 0x6d,  0x92, 0xb6, 0xdb, 0xff
]
cc48 = [
    0x00, 0x11, 0x22, 0x33,  0x44, 0x55, 0x66, 0x77,  0x88, 0x99, 0xaa, 0xbb,  0xcc, 0xdd, 0xee, 0xff
]
cc58 = [
    0x00, 0x08, 0x10, 0x19,  0x21, 0x29, 0x31, 0x3a,  0x42, 0x4a, 0x52, 0x5a,  0x63, 0x6b, 0x73, 0x7b,
    0x84, 0x8c, 0x94, 0x9c,  0xa5, 0xad, 0xb5, 0xbd,  0xc5, 0xce, 0xd6, 0xde,  0xe6, 0xef, 0xf7, 0xff
]


# https://github.com/marco-calautti/Rainbow/blob/master/Rainbow.ImgLib/ImgLib/Encoding/Implementation/ColorCodecRGB5A3.cs
def readRGB5A3(color):
    r, g, b, a = (0, 0, 0, 0)
    if color & 0x8000 != 0:
        a = 255
        r = cc58[(color >> 10) & 0x1F]
        g = cc58[(color >> 5) & 0x1F]
        b = cc58[(color) & 0x1F]
    else:
        a = cc38[(color >> 12) & 0x7]
        r = cc48[(color >> 8) & 0xF]
        g = cc48[(color >> 4) & 0xF]
        b = cc48[(color) & 0xF]
    return (r, g, b, a)


def getPaletteIndex(palette, color, fixtrasp=False, starti=0, palsize=-1, checkalpha=False, zerotrasp=True):
    if color[3] == 0 and zerotrasp:
        return 0
    if palsize == -1:
        palsize = len(palette)
    zeroalpha = -1
    for i in range(starti, starti + palsize):
        if fixtrasp and i == starti:
            continue
        if palette[i][0] == color[0] and palette[i][1] == color[1] and palette[i][2] == color[2] and (not checkalpha or palette[i][3] == color[3]):
            return i - starti
        if palette[i][3] == 0:
            zeroalpha = i - starti
    if palette[starti][0] == color[0] and palette[starti][1] == color[1] and palette[starti][2] == color[2] and (not checkalpha or palette[starti][3] == color[3]):
        return 0
    if checkalpha and color[3] == 0 and zeroalpha != -1:
        return zeroalpha
    mindist = 0xFFFFFFFF
    disti = 0
    for i in range(starti + 1, starti + palsize):
        distance = getColorDistance(color, palette[i], checkalpha)
        if distance < mindist:
            mindist = distance
            disti = i - starti
    logDebug("Color", color, "not found, closest color:", palette[disti])
    return disti


def findBestPalette(palettes, colors):
    if len(palettes) == 1:
        return 0
    mindist = 0xFFFFFFFF
    disti = 0
    for i in range(len(palettes)):
        distance = 0
        for color in colors:
            singledist = 0xFFFFFFFF
            for palcolor in palettes[i]:
                singledist = min(singledist, getColorDistance(color, palcolor))
            distance += singledist
        if distance < mindist:
            mindist = distance
            disti = i
            if mindist == 0:
                break
    return disti


def drawPalette(pixels, palette, width, ystart=0):
    for x in range(len(palette)):
        j = width + ((x % 8) * 5)
        i = ystart + ((x // 8) * 5)
        for j2 in range(5):
            for i2 in range(5):
                pixels[j + j2, i + i2] = palette[x]
    return pixels



import math
import os
from PIL import Image
from hacktools import common


# Generic extract/repack functions
def extractARC(infolder, outfolder):
    common.makeFolder(outfolder)
    common.logMessage("Extracting ARC to", outfolder, "...")
    files = common.getFiles(infolder, ".arc")
    for file in common.showProgress(files):
        common.logDebug("Processing", file, "...")
        common.execute("wszst EXTRACT " + infolder + file + " -D " + outfolder + file, False)
    common.logMessage("Done! Extracted", len(files), "files")


def extractTPL(infolder, outfolder, splitName=True):
    common.makeFolder(outfolder)
    common.logMessage("Extracting TPL to", outfolder, "...")
    files = common.getFiles(infolder, ".tpl")
    for file in common.showProgress(files):
        common.logDebug("Processing", file, "...")
        filename = file.split("/")[0] if splitName else file
        common.execute("wimgt DECODE " + infolder + file + " -D " + outfolder + filename + "/" + os.path.basename(file).replace(".tpl", ".png"), False)
    common.logMessage("Done! Extracted", len(files), "files")


def extractIso(isofile, extractfolder, workfolder=""):
    common.logMessage("Extracting ISO", isofile, "...")
    common.makeFolder(extractfolder)
    common.execute("wit EXTRACT -o {iso} {folder}".format(iso=isofile, folder=extractfolder), False)
    if workfolder != "":
        common.copyFolder(extractfolder, workfolder)
    common.logMessage("Done!")


def repackIso(isofile, isopatch, workfolder, patchfile=""):
    common.logMessage("Repacking ISO", isopatch, "...")
    if os.path.isfile(isopatch):
        os.remove(isopatch)
    common.execute("wit COPY {folder} {iso}".format(folder=workfolder, iso=isopatch), False)
    common.logMessage("Done!")


# TPL files
# http://wiki.tockdom.com/wiki/TPL_(File_Format)
class TPL:
    imgnum = 0
    tableoff = 0
    images = []


class TPLImage:
    imgoff = 0
    paloff = 0
    palformat = 0x02
    paldataoff = 0
    palette = []
    width = 0
    height = 0
    format = 0x09
    dataoff = 0
    tilewidth = 8
    tileheight = 8
    blockwidth = 0
    blockheight = 0


def readTPL(file):
    tpl = TPL()
    with common.Stream(file, "rb", False) as f:
        f.seek(4)  # Header
        tpl.imgnum = f.readUInt()
        tpl.tableoff = f.readUInt()
        tpl.images = []
        for i in range(tpl.imgnum):
            image = TPLImage()
            tpl.images.append(image)
            f.seek(tpl.tableoff + i * 8)
            image.imgoff = f.readUInt()
            image.paloff = f.readUInt()
            f.seek(image.paloff)
            palcount = f.readUShort()
            f.seek(1, 1)  # Unpacked
            f.seek(1, 1)  # Padding
            image.palformat = f.readUInt()
            image.paldataoff = f.readUInt()
            if image.palformat != 0x02:
                common.logError("Unimplemented palette format:", image.palformat)
                continue
            f.seek(image.paldataoff)
            image.palette = []
            for j in range(palcount):
                image.palette.append(common.readRGB5A3(f.readShort()))
            f.seek(image.imgoff)
            image.height = f.readUShort()
            image.width = f.readUShort()
            image.format = f.readUInt()
            image.dataoff = f.readUInt()
            if image.format != 0x08 and image.format != 0x09:
                common.logError("Unimplemented image format:", image.format)
                continue
            image.tilewidth = 8
            image.tileheight = 8 if image.format == 0x08 else 4
            image.blockwidth = math.ceil(image.width / image.tilewidth) * image.tilewidth
            image.blockheight = math.ceil(image.height / image.tileheight) * image.tileheight
    return tpl


def writeTPL(file, tpl, infile):
    with common.Stream(file, "r+b", False) as f:
        for i in range(tpl.imgnum):
            image = tpl.images[i]
            imgfile = infile
            if i > 0:
                imgfile = imgfile.replace(".png", ".mm" + str(i) + ".png")
            img = Image.open(imgfile)
            img = img.convert("RGBA")
            pixels = img.load()
            f.seek(image.dataoff)
            for y in range(0, image.blockheight, image.tileheight):
                for x in range(0, image.blockwidth, image.tilewidth):
                    for y2 in range(image.tileheight):
                        for x2 in range(image.tilewidth):
                            index = 0
                            if x + x2 < img.width and y + y2 < img.height:
                                color = pixels[x + x2, y + y2]
                                index = common.getPaletteIndex(image.palette, color, False, 0, -1, True, False)
                            if image.format == 0x08:
                                f.writeHalf(index)
                            else:
                                f.writeByte(index)


# Font files
def getFontGlyphs(file):
    glyphs = {}
    with common.Stream(file, "rb", False) as f:
        # Header
        f.seek(36)
        hdwcoffset = f.readUInt()
        pamcoffset = f.readUInt()
        common.logDebug("hdwcoffset:", hdwcoffset, "pamcoffset:", pamcoffset)
        # HDWC
        f.seek(hdwcoffset - 4)
        hdwclen = f.readUInt()
        tilenum = (hdwclen - 16) // 3
        firstcode = f.readUShort()
        lastcode = f.readUShort()
        f.seek(4, 1)
        common.logDebug("firstcode:", firstcode, "lastcode:", lastcode, "tilenum", tilenum)
        hdwc = []
        for i in range(tilenum):
            hdwcstart = f.readSByte()
            hdwcwidth = f.readByte()
            hdwclength = f.readByte()
            hdwc.append((hdwcstart, hdwcwidth, hdwclength))
        # PAMC
        nextoffset = pamcoffset
        while nextoffset != 0x00:
            f.seek(nextoffset)
            firstchar = f.readUShort()
            lastchar = f.readUShort()
            sectiontype = f.readUShort()
            f.seek(2, 1)
            nextoffset = f.readUInt()
            common.logDebug("firstchar:", common.toHex(firstchar), "lastchar:", common.toHex(lastchar), "sectiontype:", sectiontype, "nextoffset:", nextoffset)
            if sectiontype == 0:
                firstcode = f.readUShort()
                for i in range(lastchar - firstchar + 1):
                    c = common.codeToChar(firstchar + i)
                    glyphs[c] = hdwc[firstcode + i] + (firstchar + i,)
            elif sectiontype == 1:
                for i in range(lastchar - firstchar + 1):
                    charcode = f.readUShort()
                    if charcode == 0xFFFF or charcode >= len(hdwc):
                        continue
                    c = common.codeToChar(firstchar + i)
                    glyphs[c] = hdwc[charcode] + (firstchar + i,)
            else:
                common.logError("Unknown section type", sectiontype)
    return glyphs



import math
import os
import shutil
import struct
from PIL import Image, ImageOps
from psd_tools import PSDImage
from hacktools import common


# Generic extract/repack functions
def extractNSBMD(infolder, outfolder, extension=".nsbmd"):
    common.makeFolder(outfolder)
    common.logMessage("Extracting NSBMD to", outfolder, "...")
    files = common.getFiles(infolder, extension)
    for file in common.showProgress(files):
        common.logDebug("Processing", file, "...")
        nsbmd = readNSBMD(infolder + file)
        if nsbmd is not None and len(nsbmd.textures) > 0:
            common.makeFolders(outfolder + os.path.dirname(file))
            for texi in range(len(nsbmd.textures)):
                drawNSBMD(outfolder + file.replace(extension, "") + "_" + nsbmd.textures[texi].name + ".png", nsbmd, texi)
    common.logMessage("Done! Extracted", len(files), "files")


def repackNSBMD(workfolder, infolder, outfolder, extension=".nsbmd", writefunc=None):
    common.logMessage("Repacking NSBMD from", workfolder, "...")
    files = common.getFiles(infolder, extension)
    for file in common.showProgress(files):
        common.logDebug("Processing", file, "...")
        common.copyFile(infolder + file, outfolder + file)
        nsbmd = readNSBMD(infolder + file)
        if nsbmd is not None and len(nsbmd.textures) > 0:
            fixtrasp = False
            if writefunc is not None:
                fixtrasp = writefunc(file, nsbmd)
            for texi in range(len(nsbmd.textures)):
                pngname = file.replace(extension, "") + "_" + nsbmd.textures[texi].name + ".png"
                if os.path.isfile(workfolder + pngname):
                    common.logDebug(" Repacking", pngname, "...")
                    writeNSBMD(outfolder + file, nsbmd, texi, workfolder + pngname, fixtrasp)
    common.logMessage("Done!")


def extractIMG(infolder, outfolder, extensions=".NCGR", readfunc=None):
    common.makeFolder(outfolder)
    common.logMessage("Extracting IMG to", outfolder, "...")
    files = common.getFiles(infolder, extensions)
    for file in common.showProgress(files):
        common.logDebug("Processing", file, "...")
        extension = os.path.splitext(file)[1]
        if readfunc is not None:
            palettes, image, map, cell, width, height, mapfile, cellfile = readfunc(infolder, file, extension)
        else:
            palettefile = file.replace(extension, ".NCLR")
            mapfile = file.replace(extension, ".NSCR")
            cellfile = file.replace(extension, ".NCER")
            palettes, image, map, cell, width, height = readNitroGraphic(infolder + palettefile, infolder + file, infolder + mapfile, infolder + cellfile)
        if image is None:
            continue
        # Export img
        common.makeFolders(outfolder + os.path.dirname(file))
        outfile = outfolder + file.replace(extension, ".png")
        if cell is not None:
            drawNCER(outfile, cell, image, palettes, True, True)
        else:
            drawNCGR(outfile, map, image, palettes, width, height)
    common.logMessage("Done! Extracted", len(files), "files")


def repackIMG(workfolder, infolder, outfolder, extensions=".NCGR", readfunc=None, writefunc=None, clean=False):
    common.logMessage("Repacking IMG from", workfolder, "...")
    files = common.getFiles(infolder, extensions)
    for file in common.showProgress(files):
        common.logDebug("Processing", file, "...")
        extension = os.path.splitext(file)[1]
        if readfunc is not None:
            palettes, image, map, cell, width, height, mapfile, cellfile = readfunc(infolder, file, extension)
        else:
            palettefile = file.replace(extension, ".NCLR")
            mapfile = file.replace(extension, ".NSCR")
            cellfile = file.replace(extension, ".NCER")
            palettes, image, map, cell, width, height = readNitroGraphic(infolder + palettefile, infolder + file, infolder + mapfile, infolder + cellfile)
        pngfile = file.replace(extension, ".psd")
        if not os.path.isfile(workfolder + pngfile):
            pngfile = file.replace(extension, ".png")
            if not os.path.isfile(workfolder + pngfile):
                pngfile = ""
        if image is None or pngfile == "":
            if clean:
                if os.path.isfile(outfolder + file):
                    os.remove(outfolder + file)
                if os.path.isfile(outfolder + mapfile):
                    os.remove(outfolder + mapfile)
                if os.path.isfile(outfolder + cellfile):
                    os.remove(outfolder + cellfile)
            else:
                common.copyFile(infolder + file, outfolder + file)
                if os.path.isfile(infolder + mapfile):
                    common.copyFile(infolder + mapfile, outfolder + mapfile)
                if os.path.isfile(workfolder + cellfile):
                    common.copyFile(workfolder + cellfile, outfolder + cellfile)
                elif os.path.isfile(infolder + cellfile):
                    common.copyFile(infolder + cellfile, outfolder + cellfile)
            continue
        common.makeFolders(outfolder + os.path.dirname(file))
        common.copyFile(infolder + file, outfolder + file)
        transptile = False
        if writefunc is not None:
            image, map, cell, width, height, transptile = writefunc(file, image, map, cell, width, height)
        if map is None and cell is None:
            writeNCGR(outfolder + file, image, workfolder + pngfile, palettes, width, height)
        elif cell is None:
            common.copyFile(infolder + mapfile, outfolder + mapfile)
            writeMappedNSCR(outfolder + file, outfolder + mapfile, image, map, workfolder + pngfile, palettes, width, height, transptile)
        else:
            if os.path.isfile(workfolder + cellfile):
                cell = readNCER(workfolder + cellfile)
                common.copyFile(workfolder + cellfile, outfolder + cellfile)
            else:
                common.copyFile(infolder + cellfile, outfolder + cellfile)
            writeNCER(outfolder + file, outfolder + cellfile, image, cell, workfolder + pngfile, palettes, width, height)
    common.logMessage("Done!")


# Font
class FontNFTR:
    height = 0
    width = 0
    plgcoffset = 0
    hdwcoffset = 0
    pamcoffset = 0
    plgcsize = 0
    glyphwidth = 0
    glyphheight = 0
    glyphlength = 0
    depth = 0
    rotation = 0
    tilenum = 0
    firstcode = 0
    lastcode = 0
    plgc = []
    colors = []
    hdwc = []
    pamc = []
    glyphs = {}


class FontHDWC:
    start = 0
    width = 0
    length = 0


class FontPAMC:
    firstchar = 0
    lastchar = 0
    type = 0
    nextoffset = 0


class FontGlyph:
    start = 0
    width = 0
    length = 0
    char = ""
    code = 0
    index = 0

    def __init__(self, hdwc, char, code, index):
        self.start = hdwc.start
        self.width = hdwc.width
        self.length = hdwc.length
        self.char = char
        self.code = code
        self.index = index


def readNFTR(file, generateglyphs=False):
    nftr = FontNFTR()
    with common.Stream(file, "rb") as f:
        # Header
        f.seek(25)
        nftr.height = f.readByte()
        f.seek(3, 1)
        nftr.width = f.readByte()
        f.seek(2, 1)
        nftr.plgcoffset = f.readUInt()
        nftr.hdwcoffset = f.readUInt()
        nftr.pamcoffset = f.readUInt()
        # PLGC
        f.seek(nftr.plgcoffset - 4)
        nftr.plgcsize = f.readUInt()
        nftr.glyphwidth = f.readByte()
        nftr.glyphheight = f.readByte()
        nftr.glyphlength = f.readUShort()
        f.seek(2, 1)
        nftr.depth = f.readByte()
        nftr.rotation = f.readByte()
        nftr.tilenum = (nftr.plgcsize - 0x10) // nftr.glyphlength
        common.logDebug(vars(nftr))
        # Generate colors
        nftr.colors = []
        numcolors = pow(2, nftr.depth)
        for i in range(numcolors):
            nftr.colors.append((0, 0, 0, int(255 * i / (numcolors - 1))))
        # Read the glyphs graphics
        nftr.plgc = []
        if generateglyphs:
            data = f.readByte()
            for i in range(nftr.tilenum):
                glyph = Image.new("RGBA", (nftr.glyphwidth, nftr.glyphheight), nftr.colors[0])
                pixels = glyph.load()
                x = 0
                y = 0
                byteindex = 0
                bitmask = 0x80
                while byteindex < nftr.glyphlength and y < nftr.glyphheight:
                    intensitymask = pow(2, nftr.depth - 1)
                    intensity = 0
                    while intensitymask > 0:
                        if data & bitmask > 0:
                            intensity += intensitymask
                        bitmask >>= 1
                        if bitmask == 0:
                            bitmask = 0x80
                            byteindex += 1
                            data = f.readByte()
                        intensitymask >>= 1
                    if intensity > 0:
                        pixels[x, y] = nftr.colors[intensity]
                    x += 1
                    if x >= nftr.glyphwidth:
                        x = 0
                        y += 1
                while byteindex < nftr.glyphlength:
                    data = f.readByte()
                    byteindex += 1
                if nftr.rotation != 0:
                    angle = 90
                    if nftr.rotation == 2:
                        angle = 270
                    elif nftr.rotation == 3:
                        angle = 180
                    glyph = glyph.rotate(angle)
                nftr.plgc.append(glyph)
        # HDWC
        f.seek(nftr.hdwcoffset)
        nftr.firstcode = f.readUShort()
        nftr.lastcode = f.readUShort()
        f.seek(4, 1)
        nftr.hdwc = []
        for i in range(nftr.tilenum):
            hdwc = FontHDWC()
            hdwc.start = f.readSByte()
            hdwc.width = f.readByte()
            hdwc.length = f.readByte()
            common.logDebug(" ", vars(hdwc))
            nftr.hdwc.append(hdwc)
        # PAMC
        nftr.pamc = []
        nftr.glyphs = {}
        nextoffset = nftr.pamcoffset
        while nextoffset != 0x00:
            f.seek(nextoffset)
            pamc = FontPAMC()
            pamc.firstchar = f.readUShort()
            pamc.lastchar = f.readUShort()
            pamc.type = f.readUInt()
            nextoffset = pamc.nextoffset = f.readUInt()
            common.logDebug(" ", vars(pamc))
            if pamc.type == 0:
                firstcode = f.readUShort()
                for i in range(pamc.lastchar - pamc.firstchar + 1):
                    c = common.codeToChar(pamc.firstchar + i)
                    nftr.glyphs[c] = FontGlyph(nftr.hdwc[firstcode + i], c, pamc.firstchar + i, firstcode + i)
            elif pamc.type == 1:
                for i in range(pamc.lastchar - pamc.firstchar + 1):
                    charcode = f.readUShort()
                    if charcode == 0xFFFF or charcode >= len(nftr.hdwc):
                        continue
                    c = common.codeToChar(pamc.firstchar + i)
                    nftr.glyphs[c] = FontGlyph(nftr.hdwc[charcode], c, pamc.firstchar + i, charcode)
            else:
                common.logError("Unknown section type", pamc.type)
    return nftr


def extractFontData(fontfiles, out):
    if isinstance(fontfiles, str):
        fontfiles = [fontfiles]
    with common.Stream(out, "wb") as f:
        for fontfile in fontfiles:
            nftr = readNFTR(fontfile)
            for i in range(0x20, 0x7e):
                f.writeByte(nftr.glyphs[chr(i)].length)


# Graphics
class NCGR:
    width = 0
    height = 0
    bpp = 4
    tilesize = 8
    tileoffset = 0
    lineal = False
    tiles = []


class NSCR:
    width = 0
    height = 0
    maplen = 0
    mapoffset = 0
    maps = []


class Map:
    pal = 0
    xflip = False
    yflip = False
    tile = 0


class NCER:
    tbank = 0
    bankoffset = 0
    blocksize = 0
    partitionoffset = 0
    maxpartitionsize = 0
    firstpartitionoffset = 0
    banks = []


class Bank:
    cellnum = 0
    cellinfo = 0
    celloffset = 0
    objoffset = 0
    partitionoffset = 0
    partitionsize = 0
    cells = []
    xmax = 0
    ymax = 0
    xmin = 0
    ymin = 0
    width = 0
    height = 0
    layernum = 0
    duplicate = False


class Cell:
    x = 0
    y = 0
    width = 0
    height = 0
    numcell = 0
    shape = 0
    size = 0
    objoffset = 0
    tileoffset = 0
    rsflag = False
    objdisable = False
    doublesize = False
    objmode = 0
    mosaic = False
    depth = False
    xflip = False
    yflip = False
    selectparam = 0
    priority = 0
    pal = 0
    layer = -1


def readNitroGraphic(palettefile, tilefile, mapfile, cellfile):
    if not os.path.isfile(palettefile):
        common.logError("Palette", palettefile, "not found")
        return [], None, None, None, 0, 0
    palettes = readNCLR(palettefile)
    # Read tiles
    ncgr = readNCGR(tilefile)
    width = ncgr.width
    height = ncgr.height
    # Read maps
    nscr = None
    if os.path.isfile(mapfile):
        nscr = readNSCR(mapfile)
        width = nscr.width
        height = nscr.height
    # Read banks
    ncer = None
    if os.path.isfile(cellfile):
        ncer = readNCER(cellfile)
    return palettes, ncgr, nscr, ncer, width, height


def readNCLR(nclrfile):
    palettes = []
    with common.Stream(nclrfile, "rb") as f:
        # Read header
        f.seek(14)
        sections = f.readUShort()
        f.seek(20)
        length = f.readUInt()
        bpp = 8 if f.readUShort() == 0x04 else 4
        f.seek(6, 1)  # 0x00
        pallen = f.readUInt()
        if pallen == 0 or pallen > length:
            pallen = length - 0x18
        offset = f.readUInt()
        colornum = 0x10 if bpp == 4 else 0x100
        if pallen // 2 < colornum:
            colornum = pallen // 2
        common.logDebug("bpp", bpp, "length", length, "pallen", pallen, "colornum", colornum)
        # Read palettes
        f.seek(0x18 + offset)
        for i in range(pallen // (colornum * 2)):
            palette = []
            for j in range(colornum):
                palette.append(common.readPalette(f.readUShort()))
            palettes.append(palette)
        # Read index
        if sections == 2:
            f.seek(16, 1)
            indexedpalettes = {}
            for i in range(len(palettes)):
                indexedpalettes[f.readUShort()] = palettes[i]
        else:
            indexedpalettes = {i: palettes[i] for i in range(0, len(palettes))}
    common.logDebug("Loaded", len(indexedpalettes), "palettes")
    if 0 not in indexedpalettes.keys():
        indexedpalettes[0] = indexedpalettes[list(indexedpalettes.keys())[0]]
    return indexedpalettes


def readNCGR(ncgrfile):
    ncgr = NCGR()
    ncgr.tiles = []
    with common.Stream(ncgrfile, "rb") as f:
        f.seek(24)
        ncgr.height = f.readUShort()
        ncgr.width = f.readUShort()
        ncgr.bpp = 8 if f.readUInt() == 0x04 else 4
        ncgr.tilesize = 8
        f.seek(4, 1)
        flag = f.readUInt()
        ncgr.lineal = (flag & 0xFF) != 0x00
        ncgr.tilelen = f.readUInt()
        f.seek(4, 1)
        ncgr.tileoffset = f.tell()
        tiledata = f.read(ncgr.tilelen)
        if ncgr.width != 0xFFFF:
            ncgr.width *= ncgr.tilesize
            ncgr.height *= ncgr.tilesize
        common.logDebug(vars(ncgr))
        for i in range(ncgr.tilelen // (8 * ncgr.bpp)):
            singletile = []
            for j in range(ncgr.tilesize * ncgr.tilesize):
                x = i * (ncgr.tilesize * ncgr.tilesize) + j
                if ncgr.bpp == 4:
                    index = (tiledata[x // 2] >> ((x % 2) << 2)) & 0x0f
                else:
                    index = tiledata[x]
                singletile.append(index)
            ncgr.tiles.append(singletile)
    common.logDebug("Loaded", len(ncgr.tiles), "tiles")
    return ncgr


def readNSCR(nscrfile):
    nscr = NSCR()
    nscr.maps = []
    with common.Stream(nscrfile, "rb") as f:
        f.seek(24)
        nscr.width = f.readUShort()
        nscr.height = f.readUShort()
        f.seek(4, 1)
        nscr.maplen = f.readUInt()
        nscr.mapoffset = f.tell()
        mapdata = f.read(nscr.maplen)
        common.logDebug(vars(nscr))
        for i in range(0, len(mapdata), 2):
            data = struct.unpack("<h", mapdata[i:i+2])[0]
            nscr.maps.append(readMapData(data))
    common.logDebug("Loaded", len(nscr.maps), "maps")
    return nscr


def readMapData(data):
    map = Map()
    map.pal = (data >> 12) & 0xF
    map.xflip = (data >> 10) & 1
    map.yflip = (data >> 11) & 1
    map.tile = data & 0x3FF
    return map


def readNCER(ncerfile):
    ncer = NCER()
    ncer.banks = []
    with common.Stream(ncerfile, "rb") as f:
        f.seek(24)
        ncer.banknum = f.readUShort()
        ncer.tbank = f.readUShort()
        ncer.bankoffset = f.readUInt()
        ncer.blocksize = f.readUInt() & 0xFF
        ncer.partitionoffset = f.readUInt()
        for i in range(ncer.banknum):
            bank = Bank()
            bank.cells = []
            ncer.banks.append(bank)
        # Partition data
        if ncer.partitionoffset > 0:
            f.seek(16 + ncer.partitionoffset + 8)
            ncer.maxpartitionsize = f.readUInt()
            ncer.firstpartitionoffset = f.readUInt()
            f.seek(ncer.firstpartitionoffset - 8, 1)
            for i in range(ncer.banknum):
                ncer.banks[i].partitionoffset = f.readUInt()
                ncer.banks[i].partitionsize = f.readUInt()
        common.logDebug(vars(ncer))
        f.seek(16 + ncer.bankoffset + 8)
        for i in range(len(ncer.banks)):
            bank = ncer.banks[i]
            bank.cellnum = f.readUShort()
            bank.cellinfo = f.readUShort()
            bank.celloffset = f.readUInt()
            if ncer.tbank == 0x01:
                bank.xmax = f.readShort()
                bank.ymax = f.readShort()
                bank.xmin = f.readShort()
                bank.ymin = f.readShort()
                bank.width = bank.xmax - bank.xmin + 1
                bank.height = bank.ymax - bank.ymin + 1
            pos = f.tell()
            bank.objoffset = pos + (ncer.banknum - (i + 1)) * (8 if ncer.tbank == 0x00 else 0x10) + bank.celloffset
            f.seek(bank.objoffset)
            for j in range(bank.cellnum):
                cell = Cell()
                cell.objoffset = f.tell()
                obj0 = f.readUShort()
                obj1 = f.readUShort()
                obj2 = f.readUShort()
                cell.y = obj0 & 0xFF
                if cell.y >= 128:
                    cell.y -= 256
                cell.shape = (obj0 >> 14) & 3
                cell.x = obj1 & 0x01FF
                if cell.x >= 0x100:
                    cell.x -= 0x200
                cell.size = (obj1 >> 14) & 3
                cell.tileoffset = obj2 & 0x03FF
                cell.rsflag = ((obj0 >> 8) & 1) == 1
                if not cell.rsflag:
                    cell.objdisable = ((obj0 >> 9) & 1) == 1
                else:
                    cell.doublesize = ((obj0 >> 9) & 1) == 1
                cell.objmode = (obj0 >> 10) & 3
                cell.mosaic = ((obj0 >> 12) & 1) == 1
                cell.depth = ((obj0 >> 13) & 1) == 1
                if not cell.rsflag:
                    # cell.unused = (obj1 >> 9) & 7
                    cell.xflip = ((obj1 >> 12) & 1) == 1
                    cell.yflip = ((obj1 >> 13) & 1) == 1
                else:
                    cell.selectparam = (obj1 >> 9) & 0x1F
                cell.priority = (obj2 >> 10) & 3
                cell.pal = (obj2 >> 12) & 0xF
                if cell.shape == 0:
                    if cell.size == 0:
                        cellsize = (8, 8)
                    elif cell.size == 1:
                        cellsize = (16, 16)
                    elif cell.size == 2:
                        cellsize = (32, 32)
                    elif cell.size == 3:
                        cellsize = (64, 64)
                elif cell.shape == 1:
                    if cell.size == 0:
                        cellsize = (16, 8)
                    elif cell.size == 1:
                        cellsize = (32, 8)
                    elif cell.size == 2:
                        cellsize = (32, 16)
                    elif cell.size == 3:
                        cellsize = (64, 32)
                elif cell.shape == 2:
                    if cell.size == 0:
                        cellsize = (8, 16)
                    elif cell.size == 1:
                        cellsize = (8, 32)
                    elif cell.size == 2:
                        cellsize = (16, 32)
                    elif cell.size == 3:
                        cellsize = (32, 64)
                cell.width = cellsize[0]
                cell.height = cellsize[1]
                cell.numcell = j
                bank.cells.append(cell)
            # Calculate bank size
            minx = miny = 512
            maxx = maxy = -512
            malformedtbank = False
            for cell in bank.cells:
                minx = min(minx, cell.x)
                miny = min(miny, cell.y)
                maxx = max(maxx, cell.x + cell.width)
                maxy = max(maxy, cell.y + cell.height)
            if ncer.tbank == 0x01 and (maxx - minx > bank.width or maxy - miny > bank.height):
                malformedtbank = True
                common.logWarning("Malformed tbank", ncerfile, bank.width, maxx, bank.height, maxy)
            if ncer.tbank == 0x00 or malformedtbank:
                bank.width = maxx - minx
                bank.height = maxy - miny
            for cell in bank.cells:
                cell.x -= minx
                cell.y -= miny
            common.logDebug(vars(bank))
            for cell in bank.cells:
                common.logDebug(vars(cell))
            # Sort cells based on priority
            bank.cells.sort(key=lambda x: (x.priority, x.numcell), reverse=True)
            f.seek(pos)
            # Calculate layers for .psd exporting, first put the first cell on the first layer
            cells = sorted(bank.cells, key=lambda x: (x.priority, x.numcell))
            if bank.cellnum > 0:
                bank.layernum = 1
                cells[0].layer = 0
                if len(cells) > 1:
                    for j in range(1, len(cells)):
                        cell = cells[j]
                        # For every other cell in the current layer, check if it's intersected
                        hit = False
                        for layercheck in cells:
                            if cell != layercheck and layercheck.layer == bank.layernum - 1:
                                if cellIntersect(cell, layercheck):
                                    hit = True
                                    break
                        if hit:
                            # All layers are full, make a new one
                            cells[j].layer = bank.layernum
                            bank.layernum += 1
                        else:
                            cells[j].layer = bank.layernum - 1
    # Mark banks as duplicate
    for bank in ncer.banks:
        if bank.duplicate:
            continue
        for bank2 in ncer.banks:
            if bank2.duplicate or bank == bank2 or bank.cellnum != bank2.cellnum:
                continue
            samecells = True
            for i in range(bank.cellnum):
                if bank.cells[i].width != bank2.cells[i].width or bank.cells[i].height != bank2.cells[i].height or bank.cells[i].tileoffset != bank2.cells[i].tileoffset:
                    samecells = False
                    break
            if samecells:
                bank2.duplicate = True
    common.logDebug("Loaded", len(ncer.banks), "banks")
    return ncer


def cellIntersect(a, b):
    return (a.x < b.x + b.width) and (a.x + a.width > b.x) and (a.y < b.y + b.height) and (a.y + a.height > b.y)


def tileToPixels(pixels, width, ncgr, tile, i, j, palette, pali, usetrasp=True):
    for i2 in range(ncgr.tilesize):
        for j2 in range(ncgr.tilesize):
            try:
                index = ncgr.tiles[tile][i2 * ncgr.tilesize + j2]
                if not usetrasp or index > 0:
                    if ncgr.lineal:
                        lineal = (i * width * ncgr.tilesize) + (j * ncgr.tilesize * ncgr.tilesize) + (i2 * ncgr.tilesize + j2)
                        pixelx = lineal % width
                        pixely = int(math.floor(lineal / width))
                    else:
                        pixelx = j * ncgr.tilesize + j2
                        pixely = i * ncgr.tilesize + i2
                    pixels[pixelx, pixely] = palette[pali + index]
            except IndexError:
                common.logWarning("Unable to set pixels at", i, j, i2, j2, "for tile", tile, "with palette", pali)
    return pixels


def drawNCER(outfile, ncer, ncgr, palettes, usetrasp=True, layered=False):
    palsize = 0
    for palette in palettes.values():
        palsize += 5 * (len(palette) // 8)
    width = height = 0
    for bank in ncer.banks:
        if bank.duplicate:
            continue
        width = max(width, bank.width)
        height += bank.height
    img = Image.new("RGBA", (width + 40, max(height, palsize)), (0, 0, 0, 0))
    pixels = img.load()
    # Draw palette
    palstart = 0
    for palette in palettes.values():
        pixels = common.drawPalette(pixels, palette, width, palstart * 10)
        palstart += 1
    layers = []
    # If all banks have a single layer, disable layering
    if layered:
        allone = True
        for bank in ncer.banks:
            if bank.layernum > 1:
                allone = False
                break
        layered = not allone
    # Save just the palette as a separate layer
    if layered:
        img.save(outfile, "PNG")
    # Loop and draw the banks
    currheight = 0
    for bankn in range(len(ncer.banks)):
        bank = ncer.banks[bankn]
        if bank.width == 0 or bank.height == 0 or bank.duplicate:
            continue
        if layered:
            banklayers = []
            for i in range(bank.layernum):
                banklayers.append(Image.new("RGBA", (img.width, img.height), (0, 0, 0, 0)))
        for celln in range(len(bank.cells)):
            cell = bank.cells[celln]
            x = (bank.partitionoffset // (8 * ncgr.bpp)) + ((cell.tileoffset << ncer.blocksize) * 0x20 // (8 * ncgr.bpp))
            if cell.pal in palettes.keys():
                pali = 0
                palette = palettes[cell.pal]
            else:
                pali = cell.pal * 16
                palette = palettes[0]
            cellimg = Image.new("RGBA", (cell.width, cell.height), (0, 0, 0, 0))
            cellpixels = cellimg.load()
            for i in range(cell.height // ncgr.tilesize):
                for j in range(cell.width // ncgr.tilesize):
                    cellpixels = tileToPixels(cellpixels, cell.width, ncgr, x, i, j, palette, pali, usetrasp)
                    x += 1
            if cell.xflip or cell.yflip:
                if cell.yflip:
                    cellimg = ImageOps.flip(cellimg)
                if cell.xflip:
                    cellimg = ImageOps.mirror(cellimg)
            if layered:
                banklayers[cell.layer].paste(cellimg, (cell.x, currheight + cell.y), cellimg)
            img.paste(cellimg, (cell.x, currheight + cell.y), cellimg)
        if layered:
            for i in range(bank.layernum):
                layerfile = outfile.replace(".png", "_" + str(bankn) + "_" + str(i) + ".png")
                banklayers[i].save(layerfile, "PNG")
                layers.append(layerfile)
        currheight += bank.height
    if layered and shutil.which("magick"):
        with open("script.scr", "w") as script:
            script.write("\"" + outfile + "\" -label \"palette\" -background none -mosaic -set colorspace RGBA")
            for layer in layers:
                script.write(" ( -page +0+0 -label \"" + os.path.basename(layer).replace(".png", "") + "\" \"" + layer + "\"[0] -background none -mosaic -set colorspace RGBA )")
            script.write(" ( -clone 0--1 -background none -mosaic ) -reverse -write \"" + outfile.replace(".png", ".psd") + "\"")
        cmd = shutil.which("magick") + " -script script.scr"
        common.execute(cmd, False)
        for layer in layers:
            os.remove(layer)
        os.remove(outfile)
        os.remove("script.scr")
    img.save(outfile, "PNG")


def drawNCGR(outfile, nscr, ncgr, palettes, width, height, usetrasp=True):
    if width == 0xFFFF or height == 0xFFFF:
        root = int(math.sqrt(len(ncgr.tiles)))
        if math.pow(root, 2) == len(ncgr.tiles):
            width = height = root * ncgr.tilesize
            common.logWarning("Assuming square size", width, "for", outfile)
        else:
            common.logError("Wrong width/height", width, height, "for", outfile)
            return
    palsize = 0
    for palette in palettes.values():
        palsize += 5 * (len(palette) // 8)
    img = Image.new("RGBA", (width + 40, max(height, palsize)), (0, 0, 0, 0))
    pixels = img.load()
    x = 0
    for i in range(height // ncgr.tilesize):
        for j in range(width // ncgr.tilesize):
            if nscr is not None:
                map = nscr.maps[x]
                if map.pal in palettes.keys():
                    pali = 0
                    palette = palettes[map.pal]
                else:
                    pali = map.pal * 16
                    palette = palettes[0]
                pixels = tileToPixels(pixels, width, ncgr, map.tile, i, j, palette, pali, usetrasp)
                # Very inefficient way to flip pixels
                if map.xflip or map.yflip:
                    sub = img.crop(box=(j * ncgr.tilesize, i * ncgr.tilesize, j * ncgr.tilesize + ncgr.tilesize, i * ncgr.tilesize + ncgr.tilesize))
                    if map.yflip:
                        sub = ImageOps.flip(sub)
                    if map.xflip:
                        sub = ImageOps.mirror(sub)
                    img.paste(sub, box=(j * ncgr.tilesize, i * ncgr.tilesize))
            else:
                pixels = tileToPixels(pixels, width, ncgr, x, i, j, palettes[0], 0, usetrasp)
            x += 1
    palstart = 0
    for palette in palettes.values():
        pixels = common.drawPalette(pixels, palette, width, palstart * 10)
        palstart += 1
    img.save(outfile, "PNG")


def writeNCGRData(f, bpp, index1, index2):
    if bpp == 4:
        f.writeByte(((index2) << 4) | index1)
    else:
        f.writeByte(index1)
        f.writeByte(index2)


def writeNCGRTile(f, pixels, width, ncgr, i, j, palette):
    for i2 in range(ncgr.tilesize):
        for j2 in range(0, ncgr.tilesize, 2):
            if ncgr.lineal:
                lineal = (i * width * ncgr.tilesize) + (j * ncgr.tilesize * ncgr.tilesize) + (i2 * ncgr.tilesize + j2)
                pixelx = lineal % width
                pixely = int(math.floor(lineal / width))
            else:
                pixelx = j * ncgr.tilesize + j2
                pixely = i * ncgr.tilesize + i2
            index1 = common.getPaletteIndex(palette, pixels[pixelx, pixely])
            index2 = common.getPaletteIndex(palette, pixels[pixelx + 1, pixely])
            writeNCGRData(f, ncgr.bpp, index1, index2)


def writeNCGR(file, ncgr, infile, palettes, width=-1, height=-1):
    if width < 0:
        width = ncgr.width
        height = ncgr.height
    img = Image.open(infile)
    img = img.convert("RGBA")
    pixels = img.load()
    with common.Stream(file, "rb+") as f:
        f.seek(ncgr.tileoffset)
        for i in range(height // ncgr.tilesize):
            for j in range(width // ncgr.tilesize):
                writeNCGRTile(f, pixels, width, ncgr, i, j, palettes[0])


def writeNSCR(file, ncgr, nscr, infile, palettes, width=-1, height=-1):
    if width < 0:
        width = nscr.width
        # height = nscr.height
    img = Image.open(infile)
    img = img.convert("RGBA")
    pixels = img.load()
    with common.Stream(file, "rb+") as f:
        donetiles = []
        x = 0
        for i in range(height // ncgr.tilesize):
            for j in range(width // ncgr.tilesize):
                map = nscr.maps[x]
                # Skip flipped tiles since there's always(?) going to be an unflipped one next
                if map.xflip or map.yflip:
                    continue
                # Write the tile if it's a new one
                if map.tile not in donetiles:
                    donetiles.append(map.tile)
                    f.seek(ncgr.tileoffset + map.tile * (8 * ncgr.bpp))
                    writeNCGRTile(f, pixels, width, ncgr, i, j, palettes[map.pal])
                x += 1


def writeMappedNSCR(file, mapfile, ncgr, nscr, infile, palettes, width=-1, height=-1, trasnptile=False):
    if width < 0:
        width = nscr.width
        # height = nscr.height
    img = Image.open(infile)
    img = img.convert("RGBA")
    pixels = img.load()
    with common.Stream(file, "rb+") as f:
        with common.Stream(mapfile, "rb+") as mapf:
            mapf.seek(nscr.mapoffset)
            tiles = []
            if trasnptile:
                # Start with a completely transparent tile
                tile = []
                for i2 in range(ncgr.tilesize):
                    for j2 in range(ncgr.tilesize):
                        tile.append(0)
                tiles.append(tile)
                f.seek(ncgr.tileoffset)
                for i2 in range(ncgr.tilesize):
                    for j2 in range(0, ncgr.tilesize, 2):
                        writeNCGRData(f, ncgr.bpp, 0, 0)
            for i in range(height // ncgr.tilesize):
                for j in range(width // ncgr.tilesize):
                    tilecolors = []
                    for i2 in range(ncgr.tilesize):
                        for j2 in range(ncgr.tilesize):
                            tilecolors.append(pixels[j * ncgr.tilesize + j2, i * ncgr.tilesize + i2])
                    pal = common.findBestPalette(palettes, tilecolors)
                    tile = []
                    for tilecolor in tilecolors:
                        tile.append(common.getPaletteIndex(palettes[pal], tilecolor))
                    # Search for a repeated tile
                    found = -1
                    for ti in range(len(tiles)):
                        if tiles[ti] == tile:
                            found = ti
                            break
                    map = Map()
                    map.pal = pal
                    if found != -1:
                        map.tile = found
                    else:
                        tiles.append(tile)
                        map.tile = len(tiles) - 1
                        f.seek(ncgr.tileoffset + map.tile * (8 * ncgr.bpp))
                        writeNCGRTile(f, pixels, width, ncgr, i, j, palettes[map.pal])
                    mapdata = (map.pal << 12) + (map.xflip << 11) + (map.yflip << 10) + map.tile
                    mapf.writeUShort(mapdata)
            f.seek(40)
            f.writeUInt(len(tiles) * (8 * ncgr.bpp))


def writeNCER(file, ncerfile, ncgr, ncer, infile, palettes, width, height, appendTiles=False):
    psd = infile.endswith(".psd")
    if psd:
        psd = PSDImage.open(infile)
        basename = os.path.basename(infile).replace(".psd", "")
    else:
        img = Image.open(infile)
        img = img.convert("RGBA")
        pixels = img.load()
    nexttile = len(ncgr.tiles)
    with common.Stream(file, "rb+") as f:
        with common.Stream(ncerfile, "rb+") as fn:
            currheight = 0
            donetiles = []
            cellboxes = {}
            for nceri in range(len(ncer.banks)):
                bank = ncer.banks[nceri]
                if bank.width == 0 or bank.height == 0 or bank.duplicate:
                    continue
                if psd:
                    # Extract layers from the psd file, searching them by name
                    layers = []
                    for i in range(bank.layernum):
                        layername = basename + "_" + str(nceri) + "_" + str(i)
                        psdlayer = None
                        for layer in psd:
                            if layer.name == layername:
                                psdlayer = layer
                                break
                        if psdlayer is None:
                            common.logError("Layer", layername, "not found")
                            return
                        # Copy the layer in a normal PIL image for cell access, since the layer is cropped
                        layerimg = Image.new("RGBA", (psd.width, psd.height), (0, 0, 0, 0))
                        psdimg = psdlayer.topil()
                        layerimg.paste(psdimg, (psdlayer.left, psdlayer.top), psdimg)
                        layers.append(layerimg)
                for cell in bank.cells:
                    # Skip flipped cells since there's always(?) going to be an unflipped one next
                    if cell.xflip or cell.yflip:
                        continue
                    if psd:
                        img = layers[cell.layer]
                        pixels = img.load()
                    tile = (bank.partitionoffset // (8 * ncgr.bpp)) + ((cell.tileoffset << ncer.blocksize) * 0x20 // (8 * ncgr.bpp))
                    if cell.pal in palettes.keys():
                        pali = 0
                        palette = palettes[cell.pal]
                    else:
                        pali = cell.pal * 16
                        palette = palettes[0]
                    sametile = tile in donetiles
                    addingtiles = False
                    if sametile and appendTiles:
                        tiledata = []
                        for i in range(cell.height // ncgr.tilesize):
                            for j in range(cell.width // ncgr.tilesize):
                                for i2 in range(ncgr.tilesize):
                                    for j2 in range(0, ncgr.tilesize, 2):
                                        if ncgr.lineal:
                                            lineal = (i * cell.width * ncgr.tilesize) + (j * ncgr.tilesize * ncgr.tilesize) + (i2 * ncgr.tilesize + j2)
                                            pixelx = cell.x + (lineal % cell.width)
                                            pixely = currheight + cell.y + int(math.floor(lineal / cell.width))
                                        else:
                                            pixelx = cell.x + j * ncgr.tilesize + j2
                                            pixely = currheight + cell.y + i * ncgr.tilesize + i2
                                        index1 = common.getPaletteIndex(palette, pixels[pixelx, pixely], False, pali, 16 if ncgr.bpp == 4 else -1)
                                        index2 = common.getPaletteIndex(palette, pixels[pixelx + 1, pixely], False, pali, 16 if ncgr.bpp == 4 else -1)
                                        tiledata.append(index1)
                                        tiledata.append(index2)
                        sametile = tiledata == cellboxes[tile]
                        if not sametile:
                            # Check if we can find a repeated tile
                            addingtiles = True
                            tile = nexttile
                            for celltile in cellboxes:
                                if len(cellboxes[celltile]) >= len(tiledata) and tiledata == cellboxes[celltile][:len(tiledata)]:
                                    tile = celltile
                                    addingtiles = False
                                    sametile = True
                                    break
                            tileoffset = (tile * (8 * ncgr.bpp) // 0x20) >> ncer.blocksize
                            fn.seek(cell.objoffset + 4)
                            obj2 = 0
                            obj2 += tileoffset & 0x3FF
                            obj2 += (cell.priority & 3) << 10
                            obj2 += (cell.pal & 0xF) << 12
                            fn.writeUShort(obj2)
                    if not sametile:
                        currtile = tile
                        cellboxes[currtile] = []
                        for i in range(cell.height // ncgr.tilesize):
                            for j in range(cell.width // ncgr.tilesize):
                                if tile not in donetiles:
                                    donetiles.append(tile)
                                    f.seek(ncgr.tileoffset + tile * (8 * ncgr.bpp))
                                    for i2 in range(ncgr.tilesize):
                                        for j2 in range(0, ncgr.tilesize, 2):
                                            if ncgr.lineal:
                                                lineal = (i * cell.width * ncgr.tilesize) + (j * ncgr.tilesize * ncgr.tilesize) + (i2 * ncgr.tilesize + j2)
                                                pixelx = cell.x + (lineal % cell.width)
                                                pixely = currheight + cell.y + int(math.floor(lineal / cell.width))
                                            else:
                                                pixelx = cell.x + j * ncgr.tilesize + j2
                                                pixely = currheight + cell.y + i * ncgr.tilesize + i2
                                            index1 = common.getPaletteIndex(palette, pixels[pixelx, pixely], False, pali, 16 if ncgr.bpp == 4 else -1)
                                            index2 = common.getPaletteIndex(palette, pixels[pixelx + 1, pixely], False, pali, 16 if ncgr.bpp == 4 else -1)
                                            cellboxes[currtile].append(index1)
                                            cellboxes[currtile].append(index2)
                                            writeNCGRData(f, ncgr.bpp, index1, index2)
                                tile += 1
                                if addingtiles:
                                    nexttile += 1
                currheight += bank.height
        if nexttile > len(ncgr.tiles):
            tottiles = nexttile
            f.seek(32)
            f.writeUInt(tottiles)
            f.seek(4, 1)
            f.writeUInt(tottiles * (8 * ncgr.bpp))


# 3D Models
NSBMDbpp = [0, 8, 2, 4, 8, 2, 8, 16]


class NSBMD:
    textures = []
    palettes = []
    blocksize = 0
    blocklimit = 0
    texdatasize = 0
    texdataoffset = 0
    sptexsize = 0
    sptexoffset = 0
    spdataoffset = 0
    paldatasize = 0
    paldefoffset = 0
    paldataoffset = 0


class NSBMDTexture:
    name = ""
    offset = 0
    format = 0
    width = 0
    height = 0
    size = 0
    data = []
    spdata = []


class NSBMDPalette:
    name = ""
    offset = 0
    size = 0
    data = []


def readNSBMD(nsbmdfile):
    nsbmd = NSBMD()
    with common.Stream(nsbmdfile, "rb") as f:
        nsbmdstart = 0
        # 3DG have an additional header, skip it
        if f.readString(4) == "D3KT":
            f.seek(4, 1)
            nsbmdstart = f.readUShort()
        # Read the TEX0 offset
        f.seek(nsbmdstart + 20)
        check = f.readString(4)
        if check == "MDL0":
            # The model doesn't have any textures
            return None
        f.seek(-4, 1)
        texstart = f.readUShort()
        nsbmd.blockoffset = nsbmdstart + texstart
        # Read TEX0 block
        f.seek(nsbmd.blockoffset + 4)
        nsbmd.blocksize = f.readUInt()
        nsbmd.blocklimit = nsbmd.blocksize + nsbmd.blockoffset
        f.seek(4, 1)
        nsbmd.texdatasize = f.readUShort() * 8
        f.seek(6, 1)
        nsbmd.texdataoffset = f.readUInt() + nsbmd.blockoffset
        f.seek(4, 1)
        nsbmd.sptexsize = f.readUShort() * 8
        f.seek(6, 1)
        nsbmd.sptexoffset = f.readUInt() + nsbmd.blockoffset
        nsbmd.spdataoffset = f.readUInt() + nsbmd.blockoffset
        f.seek(4, 1)
        nsbmd.paldatasize = f.readUShort() * 8
        f.seek(2, 1)
        nsbmd.paldefoffset = f.readUInt() + nsbmd.blockoffset
        nsbmd.paldataoffset = f.readUInt() + nsbmd.blockoffset
        common.logDebug(vars(nsbmd))
        # Texture definition
        f.seek(1, 1)
        texnum = f.readByte()
        pos = f.tell()
        f.seek(nsbmd.paldefoffset + 1)
        palnum = f.readByte()
        f.seek(pos)
        common.logDebug("texnum:", texnum, "palnum:", palnum)
        f.seek(14 + (texnum * 4), 1)
        nsbmd.textures = []
        nsbmd.palettes = []
        for i in range(texnum):
            offset = f.readUShort() * 8
            param = f.readUShort()
            f.seek(4, 1)
            tex = NSBMDTexture()
            tex.format = (param >> 10) & 7
            tex.width = 8 << ((param >> 4) & 7)
            tex.height = 8 << ((param >> 7) & 7)
            tex.size = tex.width * tex.height * NSBMDbpp[tex.format] // 8
            if tex.format == 5:
                tex.offset = offset + nsbmd.sptexoffset
            else:
                tex.offset = offset + nsbmd.texdataoffset
            nsbmd.textures.append(tex)
        # Texture name
        for tex in nsbmd.textures:
            tex.name = f.readString(16)
            common.logDebug(vars(tex))
        # Palette definition
        f.seek(nsbmd.paldefoffset + 2 + 14 + (palnum * 4))
        for i in range(palnum):
            pal = NSBMDPalette()
            pal.offset = (f.readUShort() * 8) + nsbmd.paldataoffset
            f.seek(2, 1)
            nsbmd.palettes.append(pal)
        # Palette size
        if palnum > 0:
            for i in range(palnum):
                r = i + 1
                while r < len(nsbmd.palettes) and nsbmd.palettes[r].offset == nsbmd.palettes[i].offset:
                    r += 1
                if r != palnum:
                    nsbmd.palettes[i].size = nsbmd.palettes[r].offset - nsbmd.palettes[i].offset
                else:
                    nsbmd.palettes[i].size = nsbmd.blocklimit - nsbmd.palettes[i].offset
            nsbmd.palettes[i].size = nsbmd.blocklimit - nsbmd.palettes[i].offset
        # Palette name
        for pal in nsbmd.palettes:
            pal.name = f.readString(16)
            common.logDebug(vars(pal))
        # Traverse palettes
        for pal in nsbmd.palettes:
            f.seek(pal.offset)
            pal.data = []
            for i in range(pal.size // 2):
                pal.data.append(common.readPalette(f.readShort()))
        # Traverse texture
        spdataoffset = nsbmd.spdataoffset
        for texi in range(len(nsbmd.textures)):
            tex = nsbmd.textures[texi]
            if tex.format == 5:
                r = tex.size >> 1
                f.seek(spdataoffset)
                tex.spdata = []
                for i in range(r // 2):
                    tex.spdata.append(f.readUShort())
                spdataoffset += r
            # Export texture
            f.seek(tex.offset)
            if tex.format == 5:
                tex.data = []
                for i in range(tex.size // 4):
                    tex.data.append(f.readUInt())
            else:
                tex.data = f.read(tex.size)
        return nsbmd


def drawNSBMD(file, nsbmd, texi):
    tex = nsbmd.textures[texi]
    common.logDebug("Exporting", tex.name, "...")
    palette = None
    if tex.format != 7:
        palette = nsbmd.palettes[texi].data if texi < len(nsbmd.palettes) else nsbmd.palettes[0].data
        img = Image.new("RGBA", (tex.width + 40, max(tex.height, (len(palette) // 8) * 5)), (0, 0, 0, 0))
    else:
        img = Image.new("RGBA", (tex.width, tex.height), (0, 0, 0, 0))
    pixels = img.load()
    # A3I5 Translucent Texture (3bit Alpha, 5bit Color Index)
    if tex.format == 1:
        for i in range(tex.height):
            for j in range(tex.width):
                x = i * tex.width + j
                index = tex.data[x] & 0x1f
                alpha = (tex.data[x] >> 5) & 7
                alpha = ((alpha * 4) + (alpha // 2)) << 3
                if index < len(palette):
                    pixels[j, i] = (palette[index][0], palette[index][1], palette[index][2], alpha)
                else:
                    common.logWarning("Index", index, "is out of range", len(palette))
    # 4-color Palette
    elif tex.format == 2:
        for i in range(tex.height):
            for j in range(tex.width):
                x = i * tex.width + j
                index = (tex.data[x // 4] >> ((x % 4) << 1)) & 3
                if index < len(palette):
                    pixels[j, i] = palette[index]
                else:
                    common.logWarning("Index", index, "is out of range", len(palette))
    # 16-color Palette
    elif tex.format == 3:
        for i in range(tex.height):
            for j in range(tex.width):
                x = i * tex.width + j
                index = (tex.data[x // 2] >> ((x % 2) << 2)) & 0x0f
                if index < len(palette):
                    pixels[j, i] = palette[index]
                else:
                    common.logWarning("Index", index, "is out of range", len(palette))
    # 256-color Palette
    elif tex.format == 4:
        for i in range(tex.height):
            for j in range(tex.width):
                x = i * tex.width + j
                index = tex.data[x]
                if index < len(palette):
                    pixels[j, i] = palette[index]
                else:
                    common.logWarning("Index", index, "is out of range", len(palette))
    # 4x4-Texel Compressed Texture
    elif tex.format == 5:
        w = tex.width // 4
        h = tex.height // 4
        for y in range(h):
            for x in range(w):
                index = y * w + x
                t = tex.data[index]
                d = tex.spdata[index]
                addr = d & 0x3fff
                pali = addr << 1
                mode = (d >> 14) & 3
                for r in range(4):
                    for c in range(4):
                        texel = (t >> ((r * 4 + c) * 2)) & 3
                        i = y * 4 + r
                        j = x * 4 + c
                        try:
                            if mode == 0:
                                if texel == 3:
                                    pixels[j, i] = (0xff, 0xff, 0xff, 0)
                                else:
                                    pixels[j, i] = palette[pali + texel]
                            elif mode == 2:
                                pixels[j, i] = palette[pali + texel]
                            elif mode == 1:
                                if texel == 0 or texel == 1:
                                    pixels[j, i] = palette[pali + texel]
                                elif texel == 2:
                                    pixels[j, i] = common.sumColors(palette[pali], palette[pali + 1])
                                elif texel == 3:
                                    pixels[j, i] = (0xff, 0xff, 0xff, 0)
                            elif mode == 3:
                                if texel == 0 or texel == 1:
                                    pixels[j, i] = palette[pali + texel]
                                elif texel == 2:
                                    pixels[j, i] = common.sumColors(palette[pali], palette[pali + 1], 5, 3, 8)
                                elif texel == 3:
                                    pixels[j, i] = common.sumColors(palette[pali], palette[pali + 1], 3, 5, 8)
                        except IndexError:
                            pixels[j, i] = (0x00, 0x00, 0x00, 0xff)
    # A5I3 Translucent Texture (5bit Alpha, 3bit Color Index)
    elif tex.format == 6:
        for i in range(tex.height):
            for j in range(tex.width):
                x = i * tex.width + j
                index = tex.data[x] & 0x7
                alpha = (tex.data[x] >> 3) & 0x1f
                alpha = ((alpha * 4) + (alpha // 2)) << 3
                if index < len(palette):
                    pixels[j, i] = (palette[index][0], palette[index][1], palette[index][2], alpha)
                else:
                    common.logWarning("Index", index, "is out of range", len(palette))
    # Direct Color Texture
    elif tex.format == 7:
        for i in range(tex.height):
            for j in range(tex.width):
                x = i * tex.width + j
                p = tex.data[x * 2] + (tex.data[x * 2 + 1] << 8)
                pixels[j, i] = (((p >> 0) & 0x1f) << 3, ((p >> 5) & 0x1f) << 3, ((p >> 10) & 0x1f) << 3, 0xff if (p & 0x8000) else 0)
    # Draw palette
    if tex.format != 7:
        pixels = common.drawPalette(pixels, palette, tex.width)
    img.save(file, "PNG")


def writeNSBMD(file, nsbmd, texi, infile, fixtrasp=False):
    img = Image.open(infile)
    img = img.convert("RGBA")
    pixels = img.load()
    tex = nsbmd.textures[texi]
    with common.Stream(file, "r+b") as f:
        # Read palette
        if tex.format != 7:
            palette = nsbmd.palettes[texi]
            paldata = palette.data
        # Write new texture data
        f.seek(tex.offset)
        # A3I5 Translucent Texture (3bit Alpha, 5bit Color Index)
        if tex.format == 1:
            common.logError("Texture format 1 not implemented")
        # 4-color Palette
        elif tex.format == 2:
            for i in range(tex.height):
                for j in range(0, tex.width, 4):
                    index1 = common.getPaletteIndex(paldata, pixels[j, i], fixtrasp)
                    index2 = common.getPaletteIndex(paldata, pixels[j + 1, i], fixtrasp)
                    index3 = common.getPaletteIndex(paldata, pixels[j + 2, i], fixtrasp)
                    index4 = common.getPaletteIndex(paldata, pixels[j + 3, i], fixtrasp)
                    f.writeByte((index4 << 6) | (index3 << 4) | (index2 << 2) | index1)
        # 16/256-color Palette
        elif tex.format == 3 or tex.format == 4:
            for i in range(tex.height):
                for j in range(0, tex.width, 2):
                    index1 = common.getPaletteIndex(paldata, pixels[j, i], fixtrasp)
                    index2 = common.getPaletteIndex(paldata, pixels[j + 1, i], fixtrasp)
                    writeNCGRData(f, 4 if tex.format == 3 else 8, index1, index2)
        # 4x4-Texel Compressed Texture
        elif tex.format == 5:
            common.logError("Texture format 5 not implemented")
        # A5I3 Translucent Texture (5bit Alpha, 3bit Color Index)
        elif tex.format == 6:
            common.logError("Texture format 6 not implemented")
        # Direct Color Texture
        elif tex.format == 7:
            common.logError("Texture format 7 not implemented")


def readNitroGraphicNBFC(palettefile, tilefile, mapfile, lineal=False):
    if not os.path.isfile(palettefile):
        common.logError("Palette", palettefile, "not found")
        return [], None
    palettes = readNBFP(palettefile)
    # Read tiles
    nbfc = readNBFC(tilefile, palettes[0], lineal)
    # Read maps
    nbfs = None
    if os.path.isfile(mapfile):
        nbfs = readNBFS(mapfile)
        nbfc.width = nbfs.width
        nbfc.height = nbfs.height
    return palettes, nbfc, nbfs


def readNitroGraphicNTFT(palettefile, tilefile, lineal=True):
    if not os.path.isfile(palettefile):
        common.logError("Palette", palettefile, "not found")
        return [], None
    palettes = readNBFP(palettefile)
    # Read tiles
    ntft = readNBFC(tilefile, palettes[0], lineal)
    return palettes, ntft


def readNBFP(ntfpfile):
    indexedpalettes = {}
    palettes = []
    size = os.path.getsize(ntfpfile)
    with common.Stream(ntfpfile, "rb") as f:
        pallen = 0x200
        if size < pallen:
            pallen = size
        colornum = pallen // 2
        for i in range(size // pallen):
            palette = []
            for j in range(colornum):
                palette.append(common.readPalette(f.readUShort()))
            palettes.append(palette)
        indexedpalettes = {i: palettes[i] for i in range(0, len(palettes))}
    common.logDebug("Loaded", len(indexedpalettes), "palettes")
    return indexedpalettes


def readNBFC(ntftfile, palette, lineal):
    # RawImage image = new RawImage(file.path, file.id, TileForm.Lineal, depth, true, 0, -1, file.name);
    nbfc = NCGR()
    nbfc.tiles = []
    nbfc.bpp = 4 if len(palette) <= 16 else 8
    nbfc.width = 0x0100
    nbfc.height = 0x00C0
    nbfc.lineal = lineal
    with common.Stream(ntftfile, "rb") as f:
        tiledata = f.read()
    tilelen = len(tiledata)
    for i in range(tilelen // (8 * nbfc.bpp)):
        singletile = []
        for j in range(nbfc.tilesize * nbfc.tilesize):
            x = i * (nbfc.tilesize * nbfc.tilesize) + j
            if nbfc.bpp == 4:
                index = (tiledata[x // 2] >> ((x % 2) << 2)) & 0x0f
            else:
                index = tiledata[x]
            singletile.append(index)
        nbfc.tiles.append(singletile)
    numpix = tilelen * 8 / nbfc.bpp
    root = int(math.sqrt(numpix))
    if math.pow(root, 2) == numpix:
        nbfc.width = nbfc.height = root
        common.logDebug("Assuming square size", nbfc.width)
    else:
        nbfc.width = numpix if numpix < 0x100 else 0x0100
        nbfc.height = int(numpix // nbfc.width)
        common.logDebug("Assuming size", nbfc.width, nbfc.height)
    common.logDebug("Loaded", len(nbfc.tiles), "tiles")
    return nbfc


def readNBFS(nscrfile):
    nbfs = NSCR()
    nbfs.maps = []
    with common.Stream(nscrfile, "rb") as f:
        mapdata = f.read()
    for i in range(0, len(mapdata), 2):
        data = struct.unpack("<h", mapdata[i:i+2])[0]
        map = readMapData(data)
        map.pal = 0
        nbfs.maps.append(map)
    maplen = len(nbfs.maps)
    root = int(math.sqrt(maplen))
    if math.pow(root, 2) == maplen:
        nbfs.width = nbfs.height = root * 8
        common.logDebug("Assuming square size", nbfs.width)
    else:
        nbfs.width = 0x100 if (maplen * 8 >= 0x100) else (maplen * 8)
        nbfs.height = (maplen // (nbfs.width // 8)) * 8
        common.logDebug("Assuming size", nbfs.width, nbfs.height)
    common.logDebug("Loaded", len(nbfs.maps), "maps", nbfs.width, nbfs.height)
    return nbfs
